dnl ### Checks for programs

AC_DEFUN(P4_C_BIGENDIAN,
[AC_CACHE_CHECK(whether byte ordering is bigendian, ac_cv_c_bigendian,
[ac_cv_c_bigendian=unknown
# See if sys/param.h defines the BYTE_ORDER macro.
AC_TRY_COMPILE([#include <sys/types.h>
#include <sys/param.h>], [
#if !BYTE_ORDER || !BIG_ENDIAN || !LITTLE_ENDIAN
 bogus endian macros
#endif], [# It does; now see whether it defined to BIG_ENDIAN or not.
AC_TRY_COMPILE([#include <sys/types.h>
#include <sys/param.h>], [
#if BYTE_ORDER != BIG_ENDIAN
 not big endian
#endif], ac_cv_c_bigendian=yes, ac_cv_c_bigendian=no)])
if test $ac_cv_c_bigendian = unknown; then
AC_TRY_RUN([main () {
  /* Are we little or big endian?  From Harbison&Steele.  */
  union
  {
    long l;
    char c[sizeof (long)];
  } u;
  u.l = 1;
  exit (u.c[sizeof (long) - 1] == 1);
}], ac_cv_c_bigendian=no, ac_cv_c_bigendian=yes, 
[ echo $ac_n "(cross-compiling) ... " 2>&AC_FD_MSG ])
fi])
if test $ac_cv_c_bigendian = unknown; then
AC_MSG_CHECKING(whether byte order LOOKS bigendian)
[
cat >conftest.c <<EOF
short ascii_mm[] = { 0x4249, 0x4765, 0x6E44, 0x6961, 0x6E53, 0x7953, 0 };
short ascii_ii[] = { 0x694C, 0x5454, 0x656C, 0x6E45, 0x6944, 0x6E61, 0 };
void _ascii() { char* s = (char*)ascii_mm; s = ascii_ii); }
short ebcdic_ii[] = { 0x89D3, 0xE3E3, 0x8593, 0x95C5, 0x89C4, 0x9581, 0 };
short ebcdic_mm[] = { 0xC2C9, 0xC785, 0x95C4, 0x8981, 0x95E2, 0xA8E2, 0 };
void _ebcdic() { char* s = (char*) ebcdic_mm; s = (char*)ebcdic_ii); }
int main() { _ascii (); _ebcdic (); return 0; }
EOF
] if test -f conftest.c ; then 
     if ${CC-cc} conftest.c -o conftest.o && test -f conftest.o ; then
        if test `grep -l BIGenDianSyS conftest.o` ; then
           echo $ac_n "looks big-endian ... " 1>&AC_FD_MSG
	   ac_cv_c_bigendian=yes 
        fi
        if test `grep -l LiTTleEnDian conftest.o` ; then
           echo $ac_n 'looks little-endian ... ' 1>&AC_FD_MSG
           ac_cv_c_bigendian=no
        fi
     fi
  fi
AC_MSG_RESULT($ac_cv_c_bigendian)
fi
if test $ac_cv_c_bigendian = yes; then
  AC_DEFINE(WORDS_BIG_ENDIAN)
fi
if test $ac_cv_c_bigendian = no; then
  AC_DEFINE(WORDS_LITTLE_ENDIAN)
fi
if test $ac_cv_c_bigendian = unknown; then
  AC_MSG_ERROR(unknown endianess - sorry)
fi
])

dnl P4_SRCDIR(UNIQUE-FILE-IN-SOURCE-DIR [, USUAL-SRC-DIR])
AC_DEFUN(P4_SRCDIR,[
# A filename unique to this package, relative to the directory that
# configure is in, which we can look for to find out if srcdir is correct.
ac_unique_file=$1
ac_usual_srcdir=$2

# Find the source files, if location was not specified.
if test -z "$srcdir"; then
  ac_srcdir_defaulted=yes
  # Try the directory containing this script, then its parent.
  ac_prog=[$]0
changequote(, )dnl
  ac_confdir=`echo $ac_prog|sed 's%/[^/][^/]*$%%'`
changequote([, ])dnl
  test "x$ac_confdir" = "x$ac_prog" && ac_confdir=.
  srcdir=$ac_confdir
  if test -r $ac_confdir/$ac_unique_file; then
    srcdir=$ac_confdir
  elsif test ! -z "$ac_usual_srcdir ; then
     if test -r $ac_confdir/$ac_usual_srcdir/$ac_unique_file; then
        srcdir=$ac_confdir/$ac_usual_srcdir
     elsif test -r $ac_usual_srcdir/$ac_unique_file ; then
        srcdir=$ac_usual_srcdir 
     elseif test -r ../$ac_usual_srcdir/$ac_unique_file ; then
        srcdir=../$ac_usual_srcdir
     else
        srcdir=..
     fi
  else
     srcdir=..
  fi
else
  ac_srcdir_defaulted=no
fi
if test ! -r $srcdir/$ac_unique_file; then
  if test "$ac_srcdir_defaulted" = yes; then
    AC_MSG_ERROR(can not find sources in $ac_confdir or ..)
  else
    AC_MSG_ERROR(can not find sources in $srcdir)
  fi
fi
dnl Double slashes in pathnames in object file debugging info
dnl mess up M-x gdb in Emacs.
changequote(, )dnl
srcdir=`echo "${srcdir}" | sed 's%\([^/]\)/*$%\1%'`
changequote([, ])dnl
])


AC_DEFUN(P4_C_OLDCPP,
[AC_CACHE_CHECK(whether cpp is old-style, ac_cv_c_oldstyle_cpp,
[AC_TRY_COMPILE([#define STR(X) #X],[char* s = STR(asdf);],
ac_cv_c_oldstyle_cpp=no,
ac_cv_c_oldstyle_cpp=yes
AC_DEFINE(OLDCPP) 
)])])

dnl P4_CHECK_SYMBOL( symbol, header... [,action-if-found [, action-if-not-found]])
AC_DEFUN(P4_CHECK_SYMBOL,
[AC_MSG_CHECKING([for $1 in $2])
AC_CACHE_VAL(ac_cv_func_$1,
[AC_REQUIRE_CPP()dnl
changequote(, )dnl
symbol="[^a-zA-Z_0-9]$1[^a-zA-Z_0-9]"
changequote([, ])dnl
ac_found=no
for ac_header in $2 ; do
  ac_safe=`echo "$ac_header" | sed 'y%./+-%__p_%' `
  if test $ac_found != "yes" ; then
      if eval "test \"`echo '$ac_cv_header_'$ac_safe`\" = yes"; then
	    AC_EGREP_HEADER( $symbol, $ac_header, [ac_found="$ac_header"] )
      fi
  fi
done
if test "$ac_found" != "no" ; then 
  AC_MSG_RESULT($ac_found)
  ifelse([$3], , :, [$3])
else
  AC_MSG_RESULT(no)
  ifelse([$4], , , [$4
])dnl
fi
])])



dnl P4_CHECK_SYMBOLS( symbol..., header... [, action-if-found [, action-if-not-found]])
AC_DEFUN(P4_CHECK_SYMBOLS,
[for ac_func in $1
do
P4_CHECK_SYMBOL($ac_func, $2,
[changequote(, )dnl
  ac_tr_func=HAVE_`echo $ac_func | tr 'abcdefghijklmnopqrstuvwxyz' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'`
changequote([, ])dnl
  AC_DEFINE_UNQUOTED($ac_tr_func) $2], $3)dnl
done
])

dnl P4_CACHE_CHECK_CC_OPT(flag,cacheval)
AC_DEFUN(P4_CACHE_CHECK_CC_OPT,
[AC_CACHE_CHECK(how to pass option [$1], [$2],
[AC_SUBST($2)
echo 'void f(){}' > conftest.c
if test -z "`${CC-cc} -c $1 conftest.c 2>&1`"; then
  $2="$1"
else
  $2=""
fi
rm -f conftest*
])])

dnl P4_WITHNONE(flag,cond) - not unlike AM_CONDITIONAL
AC_DEFUN(P4_WITHNONE,
[AC_SUBST($1_WITH)
AC_SUBST($1_NONE)
if $2; then 
    $1_WITH=""
    $1_NONE="#"
else 
    $1_WITH="#"
    $1_NONE=""
fi
])

AC_DEFUN(P4_SYSDEFS,
[AC_SUBST(SYSDEFS)
changequote({, })dnl
SYSDEFS=""
tt=$target_cpu
TT=CPU_
t0=`echo "$tt" | sed -e 'y:./-:___:' `
SYSDEFS="$SYSDEFS -D$TT${t0}" 

t1=` echo "$tt" | sed -e 's:\([^0-9/.-]*[0-9]*\).*:\1:' `
t1=` echo "$t1" | sed -e 'y:./-:___:' `
if test _$t1 != _$t1 ; then 
   SYSDEFS="$SYSDEFS -D$TT$t1"
fi

t2=` echo "$tt" | sed -e 's:\([^0-9/.]*\).*:\1:' `
t2=` echo "$t2" | sed -e 'y:./-:___:' `
if test _$t0 != _$t2 -a _$t1 != _$t2 -a -n "$t2" ; then
   SYSDEFS="$SYSDEFS -D$TT$t2";
fi

tt=$target_os
TT=SYS_
echo "" >>config.h
echo '/* target   '"$tt"' */' >>config.h

t0=`echo "$tt" | sed -e 'y:./-:___:' `
SYSDEFS="$SYSDEFS -D$TT$t0" 

t1=` echo "$tt" | sed -e 's:\([^0-9/.-]*[0-9]*\).*:\1:' `
t1=` echo "$t1" | sed -e 'y:./-:___:' `
if test _$t0 != _$t1 ; then 
   SYSDEFS="$SYSDEFS -D$TT$t1" 
fi

t2=` echo "$tt" | sed -e 's:\([^0-9/.]*\).*:\1:' `
t2=` echo "$t2" | sed -e 'y:./-:___:' `
if test _$t0 != _$t2 -a _$t1 != _$t2 -a -n "$t2" ; then
   SYSDEFS="$SYSDEFS -D$TT$t2";
fi
changequote([, ])dnl
])

