/** @name dblsub.c --- Subroutines for double number (64 Bit) arithmetics.
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.25 %
 *    (%date_modified: Thu Jan 27 21:58:10 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: dblsub.c~4.25:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "dblsub.h"

void
dasl (p4dcell *a, int n)		/* left shift of *a by n positions */
{
  while (--n >= 0)
    {
      a->hi <<= 1;
      a->hi += HIGHBIT (a->lo);
      a->lo <<= 1;
    }
}

void
dasr (p4dcell *a, int n)		/* arithm. right shift of *a by n positions */
{
  while (--n >= 0)
    {
      a->lo >>= 1;
      a->lo += a->hi << (CELLBITS - 1);
      a->hi >>= 1;
    }
}
/* *INDENT-OFF* */
void
madd (p4dcell * a, p4cell b)	/* add b to a */
{
  p4ucell c;			/* carry */

  P4xD3 (*a) = c  = (p4ucell) P4xD3 (*a) + P4xW1 (b); c >>= HALFCELL;
  P4xD2 (*a) = c += (p4ucell) P4xD2 (*a) + P4xW0 (b); c >>= HALFCELL;
  P4xD1 (*a) = c += (p4ucell) P4xD1 (*a);	   c >>= HALFCELL;
  P4xD0 (*a) = c +  (p4ucell) P4xD0 (*a);
}

void
dadd (p4dcell * a, p4dcell * b)	/* add b to a */
{
  p4ucell c;			/* carry */

  P4xD3 (*a) = c  = (p4ucell) P4xD3 (*a) + P4xD3 (*b); c >>= HALFCELL;
  P4xD2 (*a) = c += (p4ucell) P4xD2 (*a) + P4xD2 (*b); c >>= HALFCELL;
  P4xD1 (*a) = c += (p4ucell) P4xD1 (*a) + P4xD1 (*b); c >>= HALFCELL;
  P4xD0 (*a) = c +  (p4ucell) P4xD0 (*a) + P4xD0 (*b);
}

void
dsub (p4dcell * a, p4dcell * b)	/* subtract b from a */
{
  p4cell c;			/* carry */

  P4xD3 (*a) = c  = (p4cell) P4xD3 (*a) - (p4cell) P4xD3 (*b); c >>= HALFCELL;
  P4xD2 (*a) = c += (p4cell) P4xD2 (*a) - (p4cell) P4xD2 (*b); c >>= HALFCELL;
  P4xD1 (*a) = c += (p4cell) P4xD1 (*a) - (p4cell) P4xD1 (*b); c >>= HALFCELL;
  P4xD0 (*a) = c +  (p4cell) P4xD0 (*a) - (p4cell) P4xD0 (*b);
}

void
dnegate (p4dcell * a)		/* negate a */
{
  p4cell c;			/* carry */

  P4xD3 (*a) = c = -(p4cell) P4xD3 (*a); c >>= HALFCELL;
  P4xD2 (*a) = c -= (p4cell) P4xD2 (*a); c >>= HALFCELL;
  P4xD1 (*a) = c -= (p4cell) P4xD1 (*a); c >>= HALFCELL;
  P4xD0 (*a) = c -  (p4cell) P4xD0 (*a);
}
/* *INDENT-ON */

int
dless (p4dcell * a, p4dcell * b)	/* result: a < b */
{
  return a->hi != b->hi
  ? a->hi < b->hi
  : a->lo < b->lo;
}

int
duless (p4udcell * a, p4udcell * b)	/* result: a < b */
{
  return a->hi != b->hi ? a->hi < b->hi : a->lo < b->lo;
}

p4udcell
ummul (p4ucell a, p4ucell b)	/* unsigned multiply, mixed precision */
{
  p4udcell res;
  p4ucell c, p;

  res.lo = (p4ucell) P4xW1 (a) * P4xW1 (b);
  if (P4xW0 (a))
    {
      p = (p4ucell) P4xW0 (a) * P4xW1 (b);
      if (P4xW0 (b))
	{
	  p4ucell q = (p4ucell) P4xW1 (a) * P4xW0 (b);
	  res.hi = (p4ucell) P4xW0 (a) * P4xW0 (b);
	  /* *INDENT-OFF* */
	  P4xD2 (res)  = c  = (p4ucell) P4xD2 (res) + P4xW1 (p) + P4xW1 (q); c >>= HALFCELL;
	  P4xD1 (res)  = c += (p4ucell) P4xD1 (res) + P4xW0 (p) + P4xW0 (q); c >>= HALFCELL;
	  /* *INDENT-ON */
	  P4xD0 (res) += c;
	}
      else
	goto three;
    }
  else
    {
      if (P4xW0 (b))
	{
	  p = (p4ucell) P4xW1 (a) * P4xW0 (b);
	three:
	  P4xD2 (res) = c = (p4ucell) P4xD2 (res) + P4xW1 (p); c >>= HALFCELL;
	  P4xD1 (res) = c + P4xW0 (p);
	  P4xD0 (res) = 0;
	}
      else
	res.hi = 0;
    }
  return res;
}

p4dcell
mmul (p4cell a, p4cell b)		/* signed multiply, mixed precision */
{
  p4dcell res;
  int s = 0;

  if (a < 0)
    a = -a, s ^= 1;
  if (b < 0)
    b = -b, s ^= 1;
  *(p4udcell *) & res = ummul (a, b);
  if (s)
    dnegate (&res);
  return res;
}

static void
shift_subtract (p4udcell * u, p4ucell v)
/* Divide unsigned double by single precision using shifts and subtracts.
   Return quotient in u->lo, remainder in u->hi. */
{
  int i = CELLBITS, c = 0;
  p4ucell q = 0, h = u->hi, l = u->lo;

  for (;;)
    {
      if (c || h >= v)
	{
	  q++;
	  h -= v;
	}
      if (--i < 0)
	break;
      c = HIGHBIT (h);
      h <<= 1;
      h += HIGHBIT (l);
      l <<= 1;
      q <<= 1;
    }
  u->hi = h;
  u->lo = q;
}

udiv_t
umdiv (p4udcell num, p4ucell denom)	/* unsigned divide procedure, mixed prec */
{
  udiv_t res;

  if (num.hi == 0)
    {
      res.quot = num.lo / denom;
      res.rem = num.lo % denom;
    }
  else
    {
      shift_subtract (&num, denom);
      res.quot = num.lo;
      res.rem = num.hi;
    }
  return res;
}

fdiv_t
smdiv (p4dcell num, p4cell denom)	/* symmetric divide procedure, mixed prec */
{
  fdiv_t res;
  int sq = 0, sr = 0;

  if (num.hi < 0)
    {
      if (num.hi == -1 && (p4cell) num.lo < 0)
	goto simple;
      dnegate (&num);
      sq ^= 1;
      sr ^= 1;
    }
  else if (num.hi == 0 && (p4cell) num.lo > 0)
    {
    simple:
      res.quot = (p4cell) num.lo / denom;
      res.rem  = (p4cell) num.lo % denom;
      return res;
    }
  if (denom < 0)
    {
      denom = -denom;
      sq ^= 1;
    }
  shift_subtract ((p4udcell *) & num, denom);
  res.quot = sq ? -num.lo : num.lo;
  res.rem  = sr ? -num.hi : num.hi;
  return res;
}

fdiv_t
fmdiv (p4dcell num, p4cell denom)	/* floored divide procedure, mixed prec */
{
  fdiv_t res = smdiv (num, denom);
  if (res.rem && (num.hi ^ denom) < 0)
    res.quot--, res.rem += denom;
  return res;
}

/*@}*/

