/** @name DOUBLE ---  The Optional Double Number Word Set
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.29 %
 *    (%date_modified: Thu Jan 27 20:06:52 2000 %)
 *
 *  @description
 *   	Double Numbers use two cells per number.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: double.c~4.29:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "support.h"
#include "dblsub.h"
#include "p4_comp.h"

#define DSP	((p4dcell *)SP)
#define UDP	((p4udcell *)SP)

FCode (p4_two_constant_RT)
{
  FX_PUSH (WP_PFA[1]);
  FX_PUSH (WP_PFA[0]);
}

FCode (p4_two_constant)
{
  p4_header (p4_two_constant_RT_ , 0);
  COMMA (FX_POP);
  COMMA (FX_POP);
}

FCode (p4_two_literal_execution)
{
  p4cell h;

  h = P4_POP (IP);
  FX_PUSH (P4_POP (IP));
  FX_PUSH (h);
}

FCode (p4_two_literal)
{
  if (STATE)
    {
      p4_compile1 ();
      COMMA (DSP->hi);
      COMMA (DSP->lo);
      SP += 2;
    }
}
P4COMPILES (p4_two_literal, p4_two_literal_execution,
	  P4_SKIPS_DCELL, P4_DEFAULT_STYLE);

FCode (p4_two_variable)
{
  p4_header (p4_create_RT_ , 0);
  COMMA (0);
  COMMA (0);
}

FCode (p4_d_plus)
{
  dadd (&DSP[1], &DSP[0]);
  SP += 2;
}

FCode (p4_d_minus)
{
  dsub (&DSP[1], &DSP[0]);
  SP += 2;
}

FCode (p4_d_dot)
{
  FX_PUSH (0);
  FX (p4_d_dot_r);
  FX (p4_space);
}

FCode (p4_d_dot_r)
{
  p4cell w = *SP++;
  int sign;

  if (*SP < 0)
    sign = 1, dnegate (&DSP[0]);
  else
    sign = 0;
  FX (p4_less_sh);
  FX (p4_sh_s);
  if (sign)
    p4_hold ('-');
  FX (p4_sh_greater);
  p4_spaces (w - *SP);
  FX (p4_type);
}

FCode (p4_d_zero_less)
{
  SP[1] = P4_FLAG (SP[0] < 0);
  SP++;
}

FCode (p4_d_zero_equals)
{
  SP[1] = P4_FLAG (SP[0] == 0 && SP[1] == 0);
  SP++;
}

FCode (p4_d_two_star)
{
  dasl ((p4dcell *) &SP[0], 1);
}

FCode (p4_d_two_slash)
{
  dasr ((p4dcell *) &SP[0], 1);
}

FCode (p4_d_less_than)
{
  SP[3] = P4_FLAG (dless (&DSP[1], &DSP[0]));
  SP += 3;
}

FCode (p4_d_to_s)
{
  SP++;
}

FCode (p4_d_equals)
{
  SP[3] = P4_FLAG (SP[2] == SP[0] && SP[3] == SP[1]);
  SP += 3;
}

FCode (p4_d_abs)
{
  if (*SP < 0)
    dnegate (&DSP[0]);
}

FCode (p4_d_max)
{
  if (dless (&DSP[1], &DSP[0]))
    DSP[1] = DSP[0];
  SP += 2;
}

FCode (p4_d_min)
{
  if (dless (&DSP[0], &DSP[1]))
    DSP[1] = DSP[0];
  SP += 2;
}

FCode (p4_d_negate)
{
  dnegate (&DSP[0]);
}

FCode (p4_m_star_slash)
{
  p4udcell lo, hi;
  p4cell p, q;
  udiv_t r1, r2;
  int sign = 0;

  if ((q = *SP++) < 0)
    q = -q, sign ^= 1;
  if ((p = *SP++) < 0)
    p = -p, sign ^= 1;
  if (*SP < 0)
    dnegate (&DSP[0]), sign ^= 1;
  hi = ummul (SP[0], p);
  lo = ummul (SP[1], p);
  madd ((p4dcell *) &hi, lo.hi);
  r1 = umdiv (hi, q);
  lo.hi = r1.rem;
  r2 = umdiv (lo, q);
  SP[0] = r1.quot;
  SP[1] = r2.quot;
  if (sign)
    dnegate (&DSP[0]);
}

FCode (p4_m_plus)
{
  madd ((p4dcell *) &SP[1], SP[0]);
  SP++;
}

FCode (p4_two_rot)
{
  p4cell h;

  h = SP[4];
  SP[4] = SP[2];
  SP[2] = SP[0];
  SP[0] = h;
  h = SP[5];
  SP[5] = SP[3];
  SP[3] = SP[1];
  SP[1] = h;
}

FCode (p4_d_u_less)
{
  SP[3] = P4_FLAG (duless (&UDP[1], &UDP[0]));
  SP += 3;
}

LISTWORDS (double) =
{
  CO ("2CONSTANT",	p4_two_constant),
  CS ("2LITERAL",	p4_two_literal),
  CO ("2VARIABLE",	p4_two_variable),
  CO ("D+",		p4_d_plus),
  CO ("D-",		p4_d_minus),
  CO ("D.",		p4_d_dot),
  CO ("D.R",		p4_d_dot_r),
  CO ("D0<",		p4_d_zero_less),
  CO ("D0=",		p4_d_zero_equals),
  CO ("D2*",		p4_d_two_star),
  CO ("D2/",		p4_d_two_slash),
  CO ("D<",		p4_d_less_than),
  CO ("D=",		p4_d_equals),
  CO ("D>S",		p4_d_to_s),
  CO ("DABS",		p4_d_abs),
  CO ("DMAX",		p4_d_max),
  CO ("DMIN",		p4_d_min),
  CO ("DNEGATE",	p4_d_negate),
  CO ("M*/",		p4_m_star_slash),
  CO ("M+",		p4_m_plus),
  CO ("2ROT",		p4_two_rot),
  CO ("DU<",		p4_d_u_less),
};
COUNTWORDS (double, "DOUBLE-EXT Double number + extensions");

/*@}*/

