/** @name helpsub.c --- subroutines for the FORTH online help utilities.
 * 
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.25 %
 *    (%date_modified: Thu Jan 27 21:54:37 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: helpsub.c~4.25:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#if defined VxWorks
#include <setjmp.h> 
#endif

#include "p4_const.h"
#include "p4options.h"
#include "help.h"


/*
 * Simple error handling: Report and abort.
 */

char *progname;			/* points in argv [0] */

void
p4_fatal (const char *msg, ...)
/*
 * quit with error message, formatting like printf()
 */
{
  char buf [0x400];
  va_list v;

  va_start (v, msg);
  vsprintf (buf, msg, v);
  va_end (v);
  fprintf (stderr, "%s: %s.\n", progname, buf);
#if defined VxWorks 
  longjmp (PFE.loop, 'X'); /*exit*/
#else
  exit(1);
#endif
}

void
p4_sys_error (void)
/*
 * quit with an error message from strerror ()
 */
{
  p4_fatal ("%s", strerror (errno));
}

void
p4_file_errorz (const char *fn)
/*
 * like sys_error when a file name is at hand.
 */
{
  p4_fatal ("File %s: %s", strerror (errno));
}

void *
p4_xalloc (size_t size)
/*
 * like calloc() but dies with erro message when failed.
 */
{
  void *p = calloc (1, size);
  if (p == NULL)
    p4_sys_error ();
  return p;
}

/*@}*/

