/* @name p4_comp.h   --- internal structure definitions 
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 4.14 %
 *    (%date_modified: Mon May 15 17:53:54 2000 %)
 *
 *  @description
 * 	lists of declarations reflecting internal structures
 *	of pfe mainly for decompiler (SEE) and debugger.
 */
/*@{*/

#ifndef __P4_COMP_H
#define __P4_COMP_H

#ifndef __P4E_H
#include "p4e.h"
#endif

#ifndef P4STRING
#ifdef OLDCPP
#define P4STR(A) "A"
#define P4CAT(A,B) A/**/B
#define P4CAT3(A,B,C) A/**/B/**/C
#define P4CAT4(A,B,C,D) A/**/B/**/C/**/D
#else
#define P4STR(A) #A
#define P4CAT(A,B) A##B
#define P4CAT3(A,B,C) A##B##C
#define P4CAT4(A,B,C,D) A##B##C##D
#endif
#define P4STRING(A) P4STR(A)
#endif

#define P4CODE(X) P4CAT(X,_)

#define  FCode(X)  void P4CODE(X) (void) /* declare a primitive */
#define P4Code(X)  void P4CODE(X) (void) /* delcare a primitive */
#define P4EX(X)   (P4CODE(X) ())

#ifdef _P4_SOURCE
# define PFX(X) P4CODE(X)
# define FX(X)  P4EX(X)
#endif

typedef struct _DecompP4 DecompP4; /* informations for the decompiler */
typedef struct _SemantP4 SemantP4; /* pointer set for state smart words */
typedef struct _Seman2P4 Seman2P4; /* dito for even smarter words like TO */

enum				/* encodings for what information */
{				/* follows the compiled word inline */
  P4_SKIPS_NOTHING,
  P4_SKIPS_OFFSET,
  P4_SKIPS_CELL,
  P4_SKIPS_DCELL,
  P4_SKIPS_FLOAT,
  P4_SKIPS_STRING,
  P4_SKIPS_2STRINGS,
  P4_SKIPS_TO_TOKEN             /* TO and +TO compile CFA or LOCAL-number */
};

struct _DecompP4		/* informations for the decompiler */
{
  unsigned skips:3;		/* this xt skips this many cells of data */
  unsigned space:3;		/* additional spaces past the word */
  unsigned cr_bef:1;		/* carriage return before printing */
  signed ind_bef:4;		/* changed indentation before print */
  unsigned cr_aft:1;		/* carriage return after print */
  signed ind_aft:4;		/* changed indentation after print */
};

struct _SemantP4		/* for words with different compilation */
{				/* and execution semantics: */
  long magic;			/* mark begin of structure */
  DecompP4 decomp;		/* decompiler aid */
  char *name;			/* compiled by */
  p4code comp;			/* compilation/interpretation semantics */
  p4code exec[1];		/* execution semantics */
};

struct _Seman2P4		/* for words with different compilation */
{				/* and two different execution semantics: */
  long magic;			/* mark begin of structure */
  DecompP4 decomp;		/* decompiler aid */
  char *name;			/* compiled by */
  p4code comp;			/* compilation/interpretation semantics */
  p4code exec[2];		/* two different execution semantics */
};				/* for cases like TO (value/local variable) */

#define P4SEMANTICS(X) P4CAT(X,_Semant)

#define P4COMPILES(C,E,S,STYLE)			\
SemantP4 P4SEMANTICS(C) =			\
{						\
  SEMANT_MAGIC,					\
  { S, STYLE },					\
  NULL,						\
  P4CODE (C),					\
  { P4CODE (E) }				\
}

#define P4COMPILES2(C,E1,E2,S,STYLE)		\
Seman2P4 P4SEMANTICS(C) =			\
{						\
  SEMANT_MAGIC,					\
  { S, STYLE },					\
  NULL,						\
  P4CODE (C),					\
  { P4CODE (E1), P4CODE (E2) }			\
}

#ifdef _P4_SOURCE
/* compile execution semantics from within C-code: */
# define COMPILE1(X)		COMMA (&P4SEMANTICS(X).exec [0])
# define COMPILE2(X)		COMMA (&P4SEMANTICS(X).exec [1])
#endif

void p4_compile1 (void);		/* compile execution semantics */
void p4_compile2 (void);		/* at system runtime */

SemantP4* p4_to_semant(p4xt xt);
char p4_category (p4code p);
void p4_decompile (char *nfa, p4xt xt);

FCode (p4_noop);
FCode (p4_ahead);

FCode (p4_backward_mark);		/* FORTH-83 style system extension words */
FCode (p4_backward_resolve);
FCode (p4_forward_mark);
FCode (p4_forward_resolve);
FCode (p4_bracket_compile);

FCode (p4_smudge);
FCode (p4_unsmudge);

/* Local variables: */

int p4_find_local (char *nm, int l);
int p4_compile_local (char *name, int len);

/* Runtimes to identify words by them: */

FCode (p4_to_execution);
FCode (p4_plus_to_execution);
FCode (p4_locals_bar_execution);
FCode (p4_locals_exit_execution);
FCode (p4_local_execution);
FCode (p4_to_local_execution);
FCode (p4_plus_to_local_execution);

FCode (p4_semicolon_execution);
FCode (p4_literal_execution);
FCode (p4_two_literal_execution);
FCode (p4_f_literal_execution); 

FCode (p4_create_RT);
FCode (p4_constant_RT);
FCode (p4_value_RT);
FCode (p4_two_constant_RT);
FCode (p4_f_constant_RT);
FCode (p4_f_variable_RT);
FCode (p4_colon_RT);
FCode (p4_debug_colon_RT);
FCode (p4_sysvar_RT);
FCode (p4_sysconst_RT);
FCode (p4_dictvar_RT);
FCode (p4_dictconst_RT);

FCode (p4_only_RT);
FCode (p4_vocabulary_RT);
FCode (p4_does_defined_RT);
FCode (p4_debug_does_defined_RT);
FCode (p4_marker_RT);
FCode (p4_defer_RT);
FCode (p4_offset_RT);

extern SemantP4 P4SEMANTICS(p4_literal);
extern SemantP4 P4SEMANTICS(p4_two_literal);
extern Seman2P4 P4SEMANTICS(p4_f_literal); 
extern Seman2P4 P4SEMANTICS(p4_semicolon);

#endif
  /* __P4_COMPILER_H */

/*@}*/

