/** @name p4_const.h ---  lots of constants and other #defines
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 4.31 %
 *    (%date_modified: Thu Jan 27 21:47:31 2000 %) 
 *
 *  @description
 *      this is most of the public part of pfe's constants 
 *
 */
/*@{*/

#ifndef __P4_CONST_H
#define	__P4_CONST_H

/*
 * Decompiler style. Each 5 comma separated numbers, meaning:
 * additional spaces after print,
 * cr before print, changed indentation for first cr,
 * cr after print, changed indentation for second cr.
 */

#define	P4_DEFAULT_STYLE	0, 0,  0,  0,  0
#define	P4_LOCALS_STYLE		0, 0,  0,  1,  0
#define	P4_SEMICOLON_STYLE	0, 0, -4,  -1, 0
#define	P4_DOES_STYLE		0, 1, -4,  0,  4

#define	P4_IF_STYLE		1, 1,  0,  0,  4
#define	P4_ELSE_STYLE		1, 1, -4,  1,  4
#define	P4_THEN_STYLE		1, 1, -4,  1,  0

#define	P4_BEGIN_STYLE		1, 1,  0,  0,  4
#define	P4_WHILE_STYLE		1, 1, -4,  0,  4
#define	P4_REPEAT_STYLE		1, 1, -4,  1,  0
#define	P4_UNTIL_STYLE		1, 1, -4,  1,  0
#define	P4_AGAIN_STYLE		1, 1, -4,  1,  0

#define	P4_DO_STYLE		1, 1,  0,  0,  4
#define	P4_LOOP_STYLE		1, 1, -4,  1,  0

#define	P4_CASE_STYLE		1, 1,  0,  0,  4
#define	P4_OF_STYLE		1, 1,  0,  0,  4
#define	P4_ENDOF_STYLE		0, 0,  0,  1, -4
#define	P4_ENDCASE_STYLE	0, 1, -4,  1,  0

/*GD* same as BEGIN, but stands alone on a line */
#define P4_BEGINS_STYLE    1, 1,  0,  1,  4

/* value style */
#define P4_NEW1_STYLE      1, 1, -2, 1, 2

/* other constants, do not change! */

#ifndef P4_VERSION
#ifdef     VERSION
#define P4_VERSION      VERSION
#else
#define	P4_VERSION	"0.28.x"
#endif
#endif

#define	P4_FALSE	((p4cell)0)
#define	P4_TRUE	        (~P4_FALSE)

#define	P4xSMUDGED	0x20	/* these bits are flags in */
#define	P4xIMMEDIATE	0x40	/* the count byte of a definition */
#define P4xONxDESTROY	0x10	/* needed only with USE_FFA */
	/* if flag is set, word being freed, shall be executed before */

#ifdef _P4_SOURCE

/* Readable macros for magic numbers suggested by spc@pineal.math.fau.edu */

#if HIGHBYTE_FIRST
#define	MAKE_MAGIC(A,B,C,D)	((p4cell)A << 24 | (p4cell)B << 16 |\
				 (p4cell)C <<  8 | (p4cell)D)
#else
#define	MAKE_MAGIC(A,B,C,D)	((p4cell)D << 24 | (p4cell)C << 16 |\
				 (p4cell)B <<  8 | (p4cell)A)
#endif

#define	SAVE_MAGIC		MAKE_MAGIC('P','F','E','S')
#define	EXCEPTION_MAGIC		MAKE_MAGIC('X','C','P','T')
#define	INPUT_MAGIC		MAKE_MAGIC('S','V','I','N')
#define	DEST_MAGIC		MAKE_MAGIC('D','E','S','T')
#define	ORIG_MAGIC		MAKE_MAGIC('O','R','I','G')
#define	LOOP_MAGIC		MAKE_MAGIC('L','O','O','P')
#define	CASE_MAGIC		MAKE_MAGIC('C','A','S','E')
#define	OF_MAGIC		MAKE_MAGIC('O','F','O','F')
#define SEMANT_MAGIC		MAKE_MAGIC('S','E','M','A')
#define P4_MAKE_MAGIC		MAKE_MAGIC('M','A','K','E')

/* THROW codes */

#define	THROW_ABORT		-1
#define	THROW_ABORT_QUOTE	-2
#define	THROW_STACK_OVER	-3
#define	THROW_STACK_UNDER	-4
#define	THROW_RSTACK_OVER	-5
#define	THROW_RSTACK_UNDER	-6
#define	THROW_INVALID_MEMORY	-9
#define	THROW_ARG_TYPE		-12
#define	THROW_UNDEFINED		-13
#define	THROW_COMPILE_ONLY	-14
#define	THROW_INVALID_FORGET	-15
#define	THROW_ZERO_NAME		-16
#define	THROW_PICNUM_OVER	-17
#define	THROW_PARSE_OVER	-18
#define	THROW_NAME_TOO_LONG	-19
#define	THROW_UNSUPPORTED	-21
#define	THROW_CONTROL_MISMATCH	-22
#define	THROW_ADDRESS_ALIGNMENT	-23
#define	THROW_USER_INTERRUPT	-28
#define	THROW_COMPILER_NESTING	-29
#define	THROW_INVALID_NAME	-32
#define	THROW_BLOCK_READ	-33
#define	THROW_BLOCK_WRITE	-34
#define	THROW_INVALID_BLOCK	-35
#define	THROW_FILE_ERROR	-37
#define	THROW_FILE_NEX		-38
#define	THROW_UNEXPECTED_EOF	-39
#define	THROW_FSTACK_OVER	-44
#define	THROW_FSTACK_UNDER	-45
#define	THROW_CURRENT_DELETED	-47
#define	THROW_SEARCH_OVER	-49
#define	THROW_SEARCH_UNDER	-50
#define	THROW_FLOATING_POINT	-55
#define	THROW_QUIT		-56
#define	THROW_IF_ELSE		-58
#define THROW_DICT_OVER         -59

#define	THROW_NO_BINARY		-2048
#define	THROW_BIN_TOO_BIG	-2049
#define	THROW_OUT_OF_MEMORY	-2050
#define	THROW_INDEX_RANGE	-2051

#endif
   /* def _P4_SOURCE */

#endif
  /* ndef __P4_CONST_H */

/*@}*/

