/** @name macros.h --- macro definitions for the portable forth environment
 * 
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see    GNU LGPL
 *  @author Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 4.31 %
 *    (%date_modified: Tue Apr 11 15:26:49 2000 %)
 */
/*@{*/

#ifndef __P4_MACRO_H
#define __P4_MACRO_H

#ifndef __P4_CONFIG_H
#include "p4_config.h"
#endif

#ifndef __P4_CONST_H
#include "p4_const.h"
#endif

#ifndef CHAR_BIT
#include <limits.h>
#endif

/* determines the dimension of any given vector */
#define DIM(X)		((int)(sizeof (X) / sizeof *(X)))
/* wipe any given vector X */
#define ZERO(X)		(memset ((X), 0, sizeof (X)))
/* copy any given vector Y to X */
#define COPY(X,Y)	(memcpy ((X), (Y), sizeof (X)))
/* return the byte offset of a given component to beginning of structure: */
#define OFFSET_OF(T,C)	((char *)&(((T *)0)->C) - (char *)0)

/* inc/decrement, push/pop of arbitrary types with arbitrary pointers */
#if defined __GNUC__ && !defined __STRICT_ANSI__ && !defined __cplusplus
/* Use non-ANSI extensions avoiding address-of operator: */
#define P4_ADD(P,N)	((char *)(P) += (N))
#define P4_INC(P,T)	(((T *)(P))++)
#define P4_DEC(P,T)	(--((T *)(P)))
#define P4_INCR(P,T,N)	(((T *)(P)) += (N))
#define P4_DECR(P,T,N)	(((T *)(P)) -= (N))
#elif defined MSDOS
#define P4_ADD(P,N)	((char *huge)(P) += (int)(N))
#define P4_INC(P,T)	(((T *huge)(P))++)
#define P4_DEC(P,T)	(--((T *huge)(P)))
#define P4_INCR(P,T,N)	(((T *huge)(P)) += (N))
#define P4_DECR(P,T,N)	(((T *huge)(P)) -= (N))
#else
/* Force (or fool) ANSI-C to do typecast's it normally refuses by */
/* casting pointers, not objects, then reference the casted pointers: */
#define P4_ADD(P,N)	(*(char **)&(P) += (N))
#define P4_INC(P,T)	((*(T **)&(P))++)
#define P4_DEC(P,T)	(--(*(T **)&(P)))
#define P4_INCR(P,T,N)	((*(T **)&(P)) += (N))
#define P4_DECR(P,T,N)	((*(T **)&(P)) -= (N))
#endif

#define P4_POPTO_(T,P,X)  ((X) = *P4_INC (P, T))
#define P4_POP_(T,P)      (*P4_INC(P,T))
#define P4_PUSH_(T,X,P)	  (*P4_DEC (P, T) = (X))
#define P4_POSH_(T,P)     (*P4_DEC (P, T))

#define P4_POPTO(P,X) 	P4_POPTO_(p4cell, P, X)
#define P4_POP(P)	P4_POP_(p4cell, P)
#define P4_PUSH(X,P)	P4_PUSH_(p4cell, X, P)
#define P4_POSH(P)      P4_POSH_(p4cell, P)

/* useful shortcuts */

#define NFA_P(X)	(p4_to_name (X))	/* CFA -> NFA */
#define CFA_P(X)	(X)			/* CFA -> CFA */
#define PFA_P(X)	(&((p4cell*)(X))[1])	/* CFA -> PFA */

#define WP_NFA		(p4_to_name(p4WP))
#define WP_CFA		(p4WP)		 
#define WP_PFA		((p4cell *)&p4WP [1]) 

#define P4CNT(X) ((unsigned)(*(p4char*)(X))) /* avoid hickups with gcc > 2.7.x */
#ifndef USE_FFA
#  define _FFA(X)	(X)		 /* NFA -> FFA w/o FFA-byte */
#  define NFACNT(X)	((X)&31) 	 /* NFA -> count of namefield */
#  define NFACNTMAX     31               /* used for buffer-sizes */
#else
#  define _FFA(X)	(&((char*)X)[-1]) /* NFA -> FFA w/ FFA-byte */
#  define NFACNT(X)	(X)
#  define NFACNTMAX	127
#endif


#define	P4_FLAG(X)	((X) ? P4_TRUE : P4_FALSE)

#define BITSOF(X)	((int)(sizeof (X) * CHAR_BIT))
#define CELLBITS	BITSOF (p4cell)
#define HALFCELL	(BITSOF (p4cell) >> 1)
#define HIGHBIT(X)	((X) >> (BITSOF (X) - 1))

#define CELL_MAX	((p4cell)((p4ucell)-1 >> 1))
#define UCELL_MAX	((p4ucell)-1)

/* #define RP		((p4cell *)p4RP) */
#define RP_PUSH(X)	P4_PUSH(X, p4RP)
#define RP_POP()	P4_POP(p4RP)
#define FX_RP           ((p4cell*)p4RP)

#if 1
#define FX_PUSH(X)      (*--p4SP = (p4cell)(X))
#define FX_POP		(*p4SP++)
#define FX_DUP		(*p4SP++ = p4SP[0])
#define FX_OVER		(*p4SP++ = p4SP[1])
#define FX_NIP		(*p4SP = *p4SP++)
#else
#define FX_PUSH(X)      (P4_PUSH((p4cell)(X), p4SP))
#define FX_POP		(P4_POP(p4SP))
#define FX_DUP		(P4_PUSH(p4SP[0], p4SP))
#define FX_OVER		(P4_PUSH(p4SP[1], p4SP))
#define FX_NIP		(*p4SP = FX_POP)
#endif
#define P4_ROOM(P,N)    (P4_DECR(P,p4cell,N))
#define P4_DROP(P,N)    (P4_INCR(P,p4cell,N))
#define FX_DROP		(P4_INC(p4SP,p4cell))
#define FX_DROP2	(P4_INCR(p4SP,p4cell,2));
#define FX_DROP3	(P4_INCR(p4SP,p4cell,3));
#define FX_DROP4	(P4_INCR(p4SP,p4cell,4));
#define FX_DROP5	(P4_INCR(p4SP,p4cell,5));
#define FX_DROP6	(P4_INCR(p4SP,p4cell,6));
#define FX_DROP7	(P4_INCR(p4SP,p4cell,7));
#define FX_2DROP	(P4_DROP(p4SP,2))
#define FX_3DROP	(P4_DROP(p4SP,3))
#define FX_4DROP	(P4_DROP(p4SP,4))
#define FX_5DROP	(P4_DROP(p4SP,5))
#define FX_6DROP	(P4_DROP(p4SP,6))
#define FX_7DROP	(P4_DROP(p4SP,7))

#if 0
#define	FX_BRANCH	P4_ADD (p4IP, *(p4cell *)p4IP)
#else
#define	FX_BRANCH	(p4IP = (p4xt*)*p4IP)
#endif
#define P4_SKIP_STRING(X)  P4_ADD (X, p4_aligned (*(p4char*)X + 1))
#define FX_SKIP_STRING     P4_SKIP_STRING(p4IP);
#define P4_ALIGNED(P)	(((size_t)(P) & (CELL_ALIGN - 1)) == 0)
#define P4_DFALIGNED(P)	(((size_t)(P) & (DFLOAT_ALIGN - 1)) == 0)
#define P4_SFALIGNED(P)	(((size_t)(P) & (SFLOAT_ALIGN - 1)) == 0)

#define P4PAD		((char *)DP + MIN_HOLD)
#define	FX_DEPTH	(PFE.s0 - p4SP)
#define	P4_COMMA(P,X)	(*(p4cell *)(P) = (p4cell)(X), P4_INC ((P), p4cell))
#define P4_FCOMMA(P,F)	(*(double *)(P) = (F), P4_INC ((P), double))
#define	FX_COMMA(X)	P4_COMMA(DP,X)
#define FX_FCOMMA(F)	P4_FCOMMA(DP,X)
#ifdef _P4_SOURCE /* these will go away */
#define	COMMA(X)	P4_COMMA(DP,X)
#define FCOMMA(X)	P4_FCOMMA(DP,X)
#endif

/* Given a p4cell or dp4cell these macros access the half cell components of it: */
/* (yes yes this is non ANSI C but it works) */
#define	P4xW0(X)		(((p4ucell_hi_lo *)&(X))->hi)
#define	P4xW1(X)		(((p4ucell_hi_lo *)&(X))->lo)

#define	P4xD0(X)		P4xW0((X).hi)
#define	P4xD1(X)		P4xW1((X).hi)
#define	P4xD2(X)		P4xW0((X).lo)
#define	P4xD3(X)		P4xW1((X).lo)

#define P4xCELL(HI,LO)	((p4cell)(HI) << HALFCELL | (LO))

#ifndef P4_NO_FP
#define P4_IF_FP_(X) X
#else
#define P4_IF_FP_(X)
#endif

#endif 
  /* __P4_MACRO_H */
/*@}*/

