/** @name types.h --- pfe's data structures
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.41 %
 *    (%date_modified: Tue Apr 11 17:10:58 2000 %)
 *
 *  @description
 *     p4th's forth data structures definitions, types and sys.names
 *
 */
/*@{*/

#ifndef __P4_TYPES_H
#define __P4_TYPES_H

#ifndef __P4_SYS_H
#include "p4sys.h"
#endif

#ifndef	_SETJMP_H
#include <setjmp.h>
#endif

#ifndef FILE
#include <stdio.h>
#endif

#if defined HAVE_SYS_TYPES_H
# include <sys/types.h>		/* size_t, time_t and friends */
#endif


#ifndef __P4OPTIONS_H
#include "p4options.h"
#endif

#include "lined.h"

/* some constants needed here, they are dependent of options.h */

#define MAX_LOCALS (1<<LD_LOCALS)
#define THREADS (1<<LD_THREADS)

#define TIB_SIZE 	0x100 /* size of terminal input buffer */
#define BPBUF		0x400 /* bytes per block */


#ifndef PATH_LENGTH		/* suggested by Andrew Houghton */

# if defined _POSIX_PATH_MAX
#   define PATH_LENGTH _POSIX_PATH_MAX
# elif defined MAXPATHLEN
#   define PATH_LENGTH MAXPATHLEN
# elif defined MAXNAMLEN
#   define PATH_LENGTH MAXNAMLEN
# elif defined PATH_MAX
#   define PATH_LENGTH PATH_MAX
# else
#   define PATH_LENGTH 256
# endif

#endif
  /* ndef PATH_LENGTH */

typedef struct Head Head;	/* pointer set comprising a definition */
typedef struct Wordl Wordl;	/* body of a word list */
typedef struct File File;	/* neccessary information about open files */
typedef struct Dict Dict;	/* a dictionary specification */
typedef struct Input Input;	/* an input source specification */
typedef struct Iframe Iframe;	/* a frame to save an input specification */
typedef struct Except Except;	/* an exception frame */

struct Head			/* set of pointers comprising a header */
  {				/* when preceeded by a name */
    char *link;			/* link back to previous word in vocabulary */
# ifndef USE_FIG
#   ifdef CPU_i960
    p4code i960_padding;        /* the i960 does align to 8-bytes anyway */
#   endif                       /* (yet each pointer is 4-bytes there ...) */
    p4code aux;			/* execution semantics or pointer for DOES> */
# endif
    p4code cfa;			/* compilation/interpretation semantics */
  };

struct Wordl			/* a word list */
  {				/* a word list identifier is a (Wordl *) */
    char *thread[THREADS];
    Wordl *prev;		/* word lists are chained */
    Wordl *also;		/* search also this dictionary, kinda FIG*/
    p4cell  flag;		/* do to_upper, a copy of sys.world_flag */
#define WORDL_NOCASE (1<<0)
#define WORDL_NOHASH (1<<1)
#define WORDL_CURRENT (1<<2)
  };

struct File			/* describes a file */
  {
    FILE *f;			/* associated ANSI-C file structure */
    char mdstr[4];		/* mode string for fopen() */
    char mode;			/* mode code for open_file() */
    signed char last_op;	/* -1 write, 0 none, 1 read */
    p4word len;			/* if stream: length of input line */
    p4ucell size;		/* if block file: size of file in blocks */
    long pos;			/* a saved position, e.g. beginning of line */
    p4ucell n;			/* block in buffer or source line */
    p4cell updated;		/* if block file: block updated? */
    char name[PATH_LENGTH];	/* file name */
    char buffer[1024];		/* buffer for block or input line */
  };

struct Input			/* an input source specification */
  {
    p4cell source_id;		/* SOURCE-ID */
    File *block_file;		/* which file is active? */
    p4ucell blk;		/* currently loaded block */
    long pos;			/* stream file: beginning of line in file */
    char *tib;			/* points to TIB or to EVALUATED string */
    p4ucell number_tib;		/* #TIB, length of string in TIB */
    p4ucell to_in;		/* input parsing position */
  };

struct Iframe			/* a frame to save an input specification */
  {
    p4cell magic;
    Input input;
    Iframe *prev;
  };

struct Except
  {
    p4cell magic;
    p4xt *ipp;
    p4cell *spp;
    p4cell *lpp;
    double *fpp;
    Iframe *iframe;
    jmp_buf jmp;
    Except *prev;
  };
  
  
typedef struct _p4_term_struct p4_term_struct;

#ifndef P4_MP_MAX
#define P4_MP_MAX 8
#endif

/* this is called options structure, since this aggregate communicates
   the command-line. But it has enhanced over time. Just expect it to
   be THE data-structure unique to each thread (started via main).
   You can actually have MP_MAX forth-processors per thread, so if
   your OS isn`t multi-threaded, what the heck, use this feature!
*/

typedef struct _p4_session* p4_sessionP;
typedef struct _p4_thread*  p4_threadP;
  
struct _p4_session
{				
        int argc;
        char** argv;
	unsigned
                canonical:2,	/* running in canonical mode */
		stdio:1,	/* standard input isn't-tty: work as filter */
	        caps_on:1,	/* exchange lower and upper case chars */
		lower_case_on:1,/* make lower case words find upper case */
		lower_case_fn:1,/* convert file names to lower case? */
		float_input:1,	/* disables floating point input when false */
		license:1,	/* show license string at startup */
		warranty:1,	/* show warranty string at startup */
		quiet:1,	/* no messages */
		verbose:1,	/* more messages */
		debug:1;	/* enable a few more outputs */
	int	cols, rows;	/* size of screen */
	p4ucell	total_size;
	p4ucell	stack_size;
	p4ucell	flt_stack_size;
	p4ucell	ret_stack_size;
	p4ucell	max_files;
	p4ucell	pockets;
	char *	p4rc_file;	/* normally "~/.pferc" */
	char *	block_file;	/* block file to use initially */
	char *	include_file;	/* file to include initially */
	char *	incpaths;
	char *	incext;
	char *	blkpaths;
	char *	blkext;
	char *	editor;		/* preferred ASCII text file editor */
        
   /* for VMs */
        int     cpus;          /* how many cpus do we have in this tread */
  
   /* additional loadlists for p4_initialize_system */
        void ** loadlist[4];   

   /* known VMs */
        int     cpu_nr;
                /* the current cpu, should be one of ->cpu[] */
        p4_threadP cpu[P4_MP_MAX];
           	/* each cpu's ->opt points back to (t)his options-struct */
                
        /* normal operations asserts: */
        /* p4TH->nr == p4TH->opt->cpu_nr */
        /* p4TH->opt->cpu[p4TH->nr] == p4TH */
        
};

#define PFE (*p4TH)

#ifndef P4_MOPTRS
#define P4_MOPTRS 128
#endif

   /* there's nothing good in this solution... *FIXME*/
#define PFE_FIG (PFE.p[125]) 

struct _p4_thread
{
/*memory*/
    void* p[P4_MOPTRS-1];
    /* p4_cleanup() will free() any pointers, including memory*
       (which is considered the first&last module chunk)
    */
    void* memory; /* the chunk that the following areas are a part of */

    p4char* dict;		/*  dictionary */
    p4char* dictlimit;
    p4cell* stack;		/*  data stack */
    p4cell* s0;
    double *fstack;		/*  floating point stack */
    double* f0;
    p4xt** rstack;		/*  return stack */
    p4xt**  r0;
    char (*pockets)[POCKET_SIZE];	/*  POCKET for interactive S" */
    char *tib;			/*  TIB */
    char* tib_end;
    char* history;		/*  command line history buffer */
    char* history_top;
    File* files;		/*  files */
    File* files_top;
#define P4_s0 PFE.s0
#define P4_f0 PFE.f0
#define P4_r0 PFE.r0

/* VM */
    p4xt* ip; /* the intruction pointer */
    p4xt  wp; /* speed up the inner interpreter */
    p4cell* sp; /* the stack pointer */
    p4xt**  rp; /* the return stack pointer */
    p4cell* lp; /* the pointer to local variables */
    double* fp; /* the floating point stack pointer */

/* jmp_buf */
    jmp_buf loop;       /* QUIT and ABORT do a THROW which longjmp() */
       			/* here thus C-stack gets cleaned up too */
/*Options*/
   int nr;                     /* this cpu's id */
   p4_sessionP set;        /* contains cpu-pointers */
#define P4_opt (*PFE.set)
    
/*Dict*/
    p4char *dp;			/* actual top of the dictionary */
    p4char *fence;		/* can't forget below that address */
    char *last;			/* NFA of most recently CREATEd header */

    Wordl *voc_link;		/* link to chained word lists */
    Wordl *context[ORDER_LEN];	/* dictionary search order */
    Wordl *only;		/* ONLY is always searched */
    Wordl *current;		/* points to vocabulary in extension */
    Wordl *dforder[ORDER_LEN];	/* default dictionary search order */
    p4xt application;		/* word to run initially or 0 */
    char *hld;			/* auxiliary pointer for number output */
    p4cell dpl;			/* position of input decimal point */

    Input input;		/* active input source, full specification */
    Input input_err;		/* input specification when error occurred */
    Iframe *saved_input;	/* links to chain of saved input specs */
    Except *cAtch;		/* links to chain of CATCHed words */
    p4ucell span;		/* char count of last EXPECT */
    p4ucell scr;		/* latest LISTed block number */
    p4cell out;			/* current output column on screen */
    p4cell state;		/* interpreting (0) or compiling (-1) */
    p4cell *locals;		/* number of locals in current def. */
    char (*local)[32];		/* names of locals in current def. */
    p4cell *csp;		/* compiler security, saves sp here */
    p4ucell base;		/* of number i/o conversion */
    p4cell precision;		/* floating point output precision */

/*Forth*/
#if defined SYS_AIX3 && !defined(__GNUC__)
    /* avoid "internal compiler error" from AIX 3.2 cc compiler */
    void *key;			/* executed by KEY */
    void *emit;			/* executed by EMIT */
    void *expect;		/* executed by EXPECT */
    void *type;			/* executed by TYPE */
#else
    p4xt key;			/* executed by KEY */
    p4xt emit;			/* executed by EMIT */
    p4xt expect;			/* executed by EXPECT */
    p4xt type;			/* executed by TYPE */
#endif
    p4cell wordl_flag;		/* do toupper() before dictionary search */
    p4cell lower_case_fn;	/* do tolower() on file names */
    p4cell redefined_msg;	/* no `"xxx" is redefined' msg if false */
    p4cell float_input;		/* don't try floating pt input when false */
    p4cell reset_order;		/* if true: reset search order on ABORT */
    p4cell pocket;		/* which pocket to use next */

    File *stdIn;		/* C-library standard files */
    File *stdOut;		/* mapped to Forth-files */
    File *stdErr;
    p4ucell more;		/* for a more-like effect */
    p4ucell lines;
    
    struct lined accept_lined;	/* better input-facilities for accep := 0*/
    p4xt  fkey_xt[10];		/* fkey_executes_xt := 0*/
    void (*execute)(p4xt);	/* := normal_execute */
    
    
/* main */
    int exitcode;
    void (*system_terminal)();
    void (*cleanup)();          /* free this thread */
    
/* dictnry */
    Wordl *atexit_wl;		/* atexit dictionary holder */
    char* volatile forget_dp;   /* temporary of forget */

/* term.h */ 
    int rows, cols;		/* size of text screen */
    int xmax, ymax;             /* size of graphics window in pixels */
    
    p4char keybuf [8];
    p4char* keyptr;
    
/* term*.c */
    void* priv;         	/* private term area, better be also in p[] */
    p4_term_struct* term; 
    char** rawkey_string; 	/* pointer to terminal escape sequences */
    char** control_string; 	/* pointer to terminal control sequences */
                        	/* as used by termunix.c */
    int (*wait_for_stdin)();
    
    void (*on_stop) (void);     /* = p4_system_terminal; */
    void (*on_continue) (void); /* = p4_interactive_terminal; */
    void (*on_winchg) (void);   /* = p4_query_winsize; */
    void (*on_sigalrm) (void);  /* really from signal.c */
    
/* debug.c */
    int debugging;
    int level;
    int maxlevel;
    long opcounter ;
    
/* yours.c */
    p4xt (*smart_char)(char c);

/* support.c/xception */
    p4code throw_cleanup;

/* make updates safer with additional padding space, use it! */
    p4cell padding[7];
};

#define DP		(PFE.dp)
#define FENCE		(PFE.fence)
#define LAST		(PFE.last)
#define VOC_LINK	(PFE.voc_link)
#define CONTEXT		(PFE.context)
#define DEFAULT_ORDER	(PFE.dforder)
#define ONLY		(PFE.only)
#define CURRENT		(PFE.current)
#define APPLICATION	(PFE.application)
#define P4HLD		(PFE.hld)
#define P4DPL		(PFE.dpl)

#define SOURCE_ID	(PFE.input.source_id)
#define SOURCE_FILE	((File*) SOURCE_ID)
#define BLOCK_FILE	(PFE.input.block_file)
#define BLK		(PFE.input.blk)
#define TIB		(PFE.input.tib)
#define NUMBER_TIB	(PFE.input.number_tib)
#define TO_IN		(PFE.input.to_in)

#define SPAN		(PFE.span)
#define SCR		(PFE.scr)
#define OUT		(PFE.out)
#define STATE		(PFE.state)
#define CSP		(PFE.csp)
#define BASE		(PFE.base)
#define PRECISION	(PFE.precision)
#define WORDL_FLAG	(PFE.wordl_flag)
#define LOWER_CASE	(PFE.wordl_flag & WORDL_NOCASE)
#define LOWER_CASE_FN	(PFE.lower_case_fn)
#define REDEFINED_MSG	(PFE.redefined_msg)
#define FLOAT_INPUT	(PFE.float_input)
#define RESET_ORDER	(PFE.reset_order)

/* -----------------------------------------------  */

enum
  {
    FMODE_RO = 1, FMODE_WO, FMODE_RW,
    FMODE_ROB, FMODE_WOB, FMODE_RWB
  };

#define FMODE_BIN (FMODE_ROB - FMODE_RO)

#if defined __IBMC__ && defined __OS2__
#error must still fix manual offset definitions *GD*
enum dictOffs			/* normally these are determined with */
  {				/* OFFSET_OF, C Set/2 can't do that. */
    do_dp		= 0,
    do_fence		= do_dp + 4,
    do_last		= do_fence + 4,
    do_voc_link		= do_last + 4,
    do_context		= do_voc_link + 4,
    do_only		= 16 + 4 * ORDER_LEN,
    do_current		= do_only + 4,
    do_forth		= do_current + 4,
    do_dforder		= do_forth + 4,
    do_application	= do_dforder + 4 * ORDER_LEN,
    do_hld		= do_application + 4,
    do_dpl		= do_hld + 4
  };
#endif

#endif
  /* ndef __P4_TYPES_H */
/*@}*/

