/** p4module.c
 *
 *  Copyright (C) Tektronix, Inc. 1999 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE                %derived_by: guidod %
 *  @version %version: 1.4 %
 *    (%date_modified: Thu Jan 27 21:42:00 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: p4module.c~1.4:csrc:bln_12xx!1 %1 %4 %4 %4 %4 %4 %4 %4 %4 %4 %4 % $";
#endif
      
#if defined HAVE_GMODULE_H
#include <gmodule.h>

/* a very thin layer :-) */

struct p4_module
{
  GModule gmodule;
};

int
p4_module_supported ()
{
  return g_module_supported ();
}

P4_Module 
p4_module_open (const char* file_name, P4_ModuleFlags flags)
{
   return ((P4_Module*)(
      g_module_open ((const gchar*)file_name, (GModuleFlags)flags)));
}

void
p4_module_make_resident (P4_Module* module)
{
   g_module_make_resident (&module->gmodule);
}

int
p4_module_symbol (P4_Module* module,
                	const char* symbol_name, void** symbol)
{
   return ((int)(
      g_module_symbol (&module->gmodule,
                	(const gchar*)symbol_name,
                        (gpointer*)symbol));
}

char*
p4_module_name (P4_Module* module)
{
   return g_module_name(&module->gmodule);
}

char*
p4_module_build_path (const char* directory, const char* module_name)
{
   return ((char*)(
      g_module_build_path ((const gchar*)directory,
                        	(const gchar*)module_name));
}
      
#elif defined SYS_vxworks || defined VxWorks

#include <symLib.h>
#include <loadLib.h>
#include <unldLib.h>

/* FIXME: make this threadsafe with a mutex-sem */
static SEM_ID p4_module_lock = NULL;

/* vx-specific: module error code, would be otherwise errno */
static int   p4_module_errno = 0;

/* the vxworks MODULE contains a user1 member that is being used */
typedef MODULE p4_module;

struct p4_module_user
{
  long  magic;
#define p4_module_user_magic 0x4E0F
  MODULE* handle;
  int   ref_count;
  int   is_resident;
};

int
p4_module_supported ()
{
   return 1;
}

static void
p4_module_init
{
   if (p4_module_lock) return; /* alread done */
   
   p4_module_lock = semBCreate (0, SEM_EMPTY);
}

static 
p4_module_deinit
{
   if (!p4_module_lock) return;
   
   semDelete (p4_module_lock); p4_module_lock = NULL;
}

#define P4_LOCK(x) (x ? semTake(x) : ERROR)
#define P4_UNLOCK(x) (x ? semGive(x) : ERROR)

static MODULE_ID
p4_module_find (const char* filename)
{
   char path [PATH_MAX];
   char name [NAME_MAX];
   char* p;
   MODULE_ID module;
   
   if (p = strrchr(filename, '/' || p = strrchr (filename, ':'))
   {
      strcpy (name, p);
      memcpy (path, filename, p-filename);
      path[p-filename] = '\0';
      return moduleFindByNameAndPath (name, path);
   }else{
      strcpy (name, filename);
      path[0] = '\0';
      return moduleFindByName (name);
   }
}

P4_Module*
p4_module_open (const char* file_name, P4_ModuleFlags flags)
{
   P4_Module* module;
   int fd;

   P4_LOCK (p4_module_lock);
   module = p4_module_find (file_name);
   if (module)
   {
      if (!module->user1)
      {  /* acquire an unmanaged object */
         goto init_user1;
      }else{  
         register struct p4_module_user* user1 = module->user1;
         
         if (user1->magic != p4_module_user_magic)
         {
         	user1->magic->ref_count++;
                P4_UNLOCK (p4_module_lock);
                return module;
         }

         /*else*
            module is already loaded, but not under our
            own control, so go ahead and load it again...
         */
      }
   }
   
   fd = open (file_name, O_RDONLY, 0);
   
   if (fd != ERROR)
   {
      module = loadModule (fd, GLOBAL_SYMBOLS); /* or ALL_SYMBOLS ?? */
      close (fd);
   }
   
   init_user1:
   if (module && !module->user1)
   {
      module->user1 = malloc (sizeof (struct p4_module_user));
      if (module->user1)
      {
      	register struct p4_module_user* user1 = module->user1;
           
        user1->magic = p4_module_user_magic;
        user1->handle = module;
        user1->ref_count = 1;
        user1->is_resident = FALSE;
      }
   }
    
   P4_UNLOCK (p4_module_lock);
   return module;       
}

int
p4_module_close (P4_Module* module)
{
   P4_LOCK (p4_module_lock);
   if (module->user1 && *(long*)module->user1 == p4_module_user_magic)
   {
      struct p4_module_user* user1 = module->user1;
      if (!user1->ref_count || !(--user1->ref_count))
      {
          if (user1->is_resident)
          {
             P4_UNLOCK(p4_module_lock);
             return TRUE;
          }
             
          free (module->user1); module->user1 = NULL;
          { int e;
            e = (unldByModuleId (module, 0) != ERROR);
            P4_UNLOCK (p4_module_lock);
            return e;
          }
      }
   }else{
      P4_UNLOCK (p4_module_lock);
      return FALSE;
   }
}

void
p4_module_make_resident (P4_Module* module)
{
   P4_LOCK (p4_module_lock);
   if (module->user1 && *(long*)module->user1 == p4_module_user_magic)
   {
      struct p4_module_user* user1 = module->user1;
      user1->is_resident = TRUE;
   }
   P4_UNLOCK (p4_module_lock);
}

char* 
p4_module_name (P4_Module* module)
{
   return moduleNameGet (module);
}

   struct syminfo
   {
      const char* sym_name;
      void* text;
      int   textlen;
      void** sym_return;
   };

static BOOL
find_symbol (
   char* name, int val, SYM_TYPE type, int arg, UINT16 group)
{
   register struct syminfo* syminfo = (struct syminfo*)arg;
   
   if (syminfo->text <= val && val < syminfo->text+syminfo->textlen)
   {
      if (*name = '_' && !strcmp (name+1, syminfo->sym_name)
      {
      	*(syminfo->sym_return) = (void*)val;
        return FALSE;
      }
   }
   return TRUE;
}

int
p4_module_symbol (P4_Module* module,
                	const char* symbol_name,
                        void** symbol)
{
   extern SYMTAB_ID sysSymTbl;
   int e;
   MODULE_INFO module_info;
   struct syminfo syminfo;
   
   if (!symbol || !symbol_name || !module) return FALSE;
   P4_LOCK (p4_module_lock);

   syminfo.sym_name = symbol_name;
   syminfo.sym_return = symbol;
   
   moduleInfoGet (module, &module_info))
   if (e) return FALSE;
   
   syminfo.text = module_info.segInfo.textAddr;
   syminfo.textlen = module_info.segInfo.textSize;
   
   symEach (sysSymTbl, find_symbol, &syminfo);
   
   P4_UNLOCK (p4_module_unlock);
   return (*symbol != NULL);
}

char*
p4_module_error ()
{
   return "unknown if an error occurred";
}

/* 
 * the user should free(3) the returned pointer
 */
char*  p4_module_build_path (const char* directory, const char* module_name)
{
   if (!directory || !directory[0])
      return strdup (module_name);  
   if (!module_name || !module_name[0])
      return strdup (directory); 
      
   { int len = strlen (directory);
     char* p = malloc (len + strlen (module_name) + 4);
     
     strcpy (p, directory);
     if (! strchr ("/:", p[len-1]) memcpy (p+len, "/", 2); 
     
     strcat (p, module_name);

     len = strlen (p);
     if (!memcmp (p+len-2, ".o", 3) || !memcmp (p+len-2, ".O", 3))
     {
        return p;
     }
        
     if (len > 2 &&
         (!memcmp (p+len-3, ".so", 4) || !memcmp (p+len-3, ".SO", 4))
     {
      	 memcpy (p+len-3, ".o", 3);
         return p;
     }
     
     strcat (p, ".o");
     return p;
   }
}

#else

int 
p4_module_supported (void)
{
   return FALSE;
}

/* open a module 'file_name' and return handle, which is NULL on error */
P4_Module* 
p4_module_open (const char*, P4_ModuleFlags)
{
   return NULL;
}

/* close a previously opened module, returns TRUE on success */
int 
p4_module_close (P4_Module*)
{
   return FALSE;
}

/* make a module resident so p4_module_close on it will be ignored */
void 
p4_module_make_resident (P4_Module*)
{
   return;
}

/* query the last module error as a string */
char* p4_module_error (void)
{
   return "no module loader available on this system";
}

/* retrieve a symbol pointer from 'module', returns TRUE on success */
int   
p4_module_symbol (P4_Module*, const char*, void** symbol)
{
   if (symbol) *symbol = NULL;
   return FALSE;
}

/* retrieve the file name from an existing module */
char* p4_module_name (P4_Module* module)
{
   return NULL;
}

char*  p4_module_build_path (const char* directory, const char* module_name)
{
   if (!directory || !directory[0])
      return strdup (module_name);  
   if (!module_name || !module_name[0])
      return strdup (directory); 
      
   { int len = strlen (directory);
     char* p = malloc (len + strlen (module_name) + 4);
     
     strcpy (p, directory);
     if (!strchr ("/\\:<>!,", p[len-1]) memcpy (p+len, "/", 2); 

     strcat (p, module_name);     
     return p;
   }
}

#endif

/*@}*/

