/** @name p4module.h
 *
 *  Copyright (C) Tektronix, Inc. 1999 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE                %derived_by: guidod %
 *  @version %version: 1.1 %
 *    (%date_modified: Thu Jan 27 21:40:24 2000 %)
 *    (%full_filespec: p4module.h~1.1:incl:bln_12xx!1 %) 
 *
 *  @description
 *            implementing the module loading mechanism,
 *            this file resembles the gmodule.h from glib,
 *            and not too accidently...
 *
 */

#ifndef __P4_MODULE_H
#define __P4_MODULE_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef enum
{
  P4_MODULE_BIND_LAZY = 1 << 0;
  P4_MODULE_BIND_MASK = 0x01;
} P4_ModuleFlags;

typedef struct p4_module P4_Module;

/* return TRUE if dynamic module loading is supported */
int p4_module_supported (void);

/* open a module 'file_name' and return handle, which is NULL on error */
P4_Module* p4_module_open (const char* file_name, P4_ModuleFlags flags);

/* close a previously opened module, returns TRUE on success */
int p4_module_close (P4_Module* module);

/* make a module resident so p4_module_close on it will be ignored */
void p4_module_make_resident (P4_Module* module);

/* query the last module error as a string */
char* p4_module_error (void);

/* retrieve a symbol pointer from 'module', returns TRUE on success */
int   p4_module_symbol (P4_Module* module, 
                        	const char* symbol_name, void** symbol);

/* retrieve the file name from an existing module */
char* p4_module_name (P4_Module* module);

/* Build the actual file name containing a module. `directory' is the
 * directory where the module file is supposed to be, or NULL or empty
 * in which case it should either be in the current directory or, on
 * some operating systems, in some standard place, for instance on the
 * PATH. Hence, to be absoultely sure to get the correct module,
 * always pass in a directory. The file name consists of the directory,
 * if supplied, and `module_name' suitably decorated accoring to
 * the operating system's conventions (for instance lib*.so or *.dll).
 *
 * No checks are made that the file exists, or is of correct type.
 */
char*  p4_module_build_path (const char* directory, const char* module_name);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __P4_MODULE_H */

