/** @name p4options.h ---	user configurable options
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE            @(#) %derived_by: guidod %
 *  @version %version: 4.28 %
 *    (%date_modified: Tue Apr 11 15:33:06 2000 %)
 *
 *  @description
 *                      User configurable options, change to suit your needs.
 *			See file `install' for explanations.
 */
/*@{*/

#ifndef __P4_OPTIONS_H
#define __P4_OPTIONS_H

/*
 * options for MFORTH at Tektronix CTE Berlin GmbH
 */
#define MFORTH 1

#ifdef MFORTH			/* MFORTH semantics */
#ifndef P4_C_QUOTE
#define P4_C_QUOTE 1		/* quote means c_quote */
#endif
#ifndef USE_NO_FFA
# ifndef USE_FFA
# define USE_FFA 1		/* use seperate FlagField */
# endif
#endif
#ifdef USE_DOLLARHEX
#undef USE_DOLLARHEX 		/* no $10FE available */
#endif
#if 0 /* there's a bug in the openstring routine */
#ifndef P4_USE_openstring
#define P4_USE_openstring 1     /* openforth's semantics of quoted strings */
#endif
#endif
#ifndef P4_MODULES
#define P4_MODULES 1		/* use shared object modules */
#endif			
#ifndef P4_STDC
#define P4_STDC 1		/* some words from the C-language family */
#endif
#endif

/*
 * System location on disk and file name conventions:
 */

#if defined DOS_FILENAMES	/* MS-DOS like file and path names */

#if defined SYS_WATCOM
# define _STR_(X) #X
#else
# define _STR_(X) P4STRING(X)
#endif

# ifndef EPREFIX
#   ifdef PREFIX
#   define EPREFIX		_STR_(PRFIX)
#   else
#   define EPREFIX		"C:"
#   endif
# endif
# ifndef LIBDIR
#   ifdef PFELIB
#   define LIBDIR		PFELIB
#   else
#   define LIBDIR		EPREFIX"\\PFE"
#   endif
# endif
# ifdef PFEHLP
#   define HELPDIR		PFEHLP
# else
#   define HELPDIR		LIBDIR"\\HELP"
# endif

# define PATH_DELIMITER		';'
# define DIR_DELIMITER		'\\'
# define DIR_DELIMSTR           "\\"
# define INC_PATHS		";"LIBDIR
# define INC_EXTENSIONS		";.P4;.FS;.4TH;.FTH"
# define BLK_PATHS		INC_PATHS
# define BLK_EXTENSIONS		";.BLK;.SCR;.FB"
# define P4RC_FILE		"PFE0.RC"
# define DEFAULT_BLKFILE	"PFE.BLK"
# define LLCMD			"DIR"
# define LSCMD			"DIR /W"

#elif defined UNIX_FILENAMES	/* UNIX-like file and path names */

# ifndef EPREFIX
#   ifdef PREFIX		
#   define EPREFIX		_STR_(PREFIX)
#   else
#   define EPREFIX		"/usr/local"
#   endif
# endif
# ifndef LIBDIR
#   ifdef PFELIB
#   define LIBDIR		PFELIB
#   else
#   define LIBDIR		EPREFIX"/lib/pfe"
#   endif
# endif
# ifdef PFEHLP
#   define HELPDIR		PFEHLP
# else
#   define HELPDIR		LIBDIR"/help"
# endif

# define PATH_DELIMITER		':'
# define DIR_DELIMITER		'/'
# define DIR_DELIMSTR           "/"
# define INC_PATHS		":~/.pfe:"LIBDIR
# define INC_EXTENSIONS		":.p4:.fs:.4th:.fth:.forth"
# define BLK_PATHS		INC_PATHS
# define BLK_EXTENSIONS		":.blk:.scr:.fb"
# define P4RC_FILE		"~/.pferc"
# define DEFAULT_BLKFILE	"pfe.blk"
# define LLCMD			"ls -alF"
# define LSCMD			"ls -C"

#elif defined URL_FILENAMES	/* WEB-like file and path names */

# ifndef EPREFIX
#   ifdef PREFIX		
#   define EPREFIX		_STR_(PREFIX)
#   else
#   define EPREFIX		"file:/usr/local"
#   endif
# endif
# ifndef LIBDIR
#   ifdef PFELIB
#   define LIBDIR	        PFELIB
#   else
#   define LIBDIR		EPREFIX"/lib/pfe"
#   endif
# endif
# ifdef PFEHLP
#   define HELPDIR		PFEHLP
# else
#   define HELPDIR		LIBDIR"/help"
# endif

# define PATH_DELIMITER		';'
# define DIR_DELIMITER		'/'
# define DIR_DELIMSTR           "/"
# define INC_PATHS		";~/.pfe;"LIBDIR
# define INC_EXTENSIONS		";.p4;.fs;.4th;.fth;.forth"
# define BLK_PATHS		INC_PATHS
# define BLK_EXTENSIONS		";.blk;.scr;.fb"
# define P4RC_FILE		"~/.pferc"
# define DEFAULT_BLKFILE	"pfe.blk"
# define LLCMD			"ls -alF"
# define LSCMD			"ls -C"

#else

#error "Don't know what kind of file names your system uses, check pfe_config.h"

#endif


/*
 * System size in memory:
 */
 
#ifndef P4_KB
#define P4_KB 1024	/* size of system altogether in KBytes */
#endif
#define TOTAL_SIZE	(P4_KB * 1024)
#define	STACK_SIZE	0	/* 0 -> TOTAL_SIZE / 8 */
#define	FLT_STACK_SIZE	0	/* 0 -> TOTAL_SIZE / 16 */
#define	RET_STACK_SIZE	0	/* 0 -> TOTAL_SIZE / 16 */
				/* the rest is used for dictionary, history, */
				/* string and block buffers, files etc. */

/*
 * Number and size of various system buffers:
 */

#define	MAX_FILES	0x10	/* maximum number of open files */
#define POCKETS		8	/* number of buffers and */
#define	POCKET_SIZE	0x100	/* size of buffers for interpretation of S" */

#define	MIN_HOLD	0x100	/* minimum space for <# # #S HOLD #> etc. */
#define	MIN_PAD		0x400	/* minimum free space in PAD */
#define	HISTORY_SIZE	0x1000	/* size of command line history buffer */


/*
 * Options controlling the dictionary structure:
 */

#define	ORDER_LEN	16	/* maximum word lists in search order */
#define	LD_THREADS	5	/* 2^n is number of threads in a words list */
#define	LD_LOCALS	5	/* 2^n is number of local variables allowed */


/*
 * Options controlling input and output:
 */

#define	TEXT_COLS	80	/* used only in case p4th fails determining */
#define	TEXT_ROWS	25	/* the screen size */

#ifdef  USE_DOLLARHEX
#define	PREFIX_HEX	'$'	/* 0 or prefix for input of hex numbers */
#define	PREFIX_BINARY	'%'	/* 0 or prefix for input of binary numbers */
#else
#define PREFIX_HEX      0
#define PREFIX_BINARY   0
#endif
#define PREFIX_0x       1       /* 0x10 =16, 0X100 = 256 */
#define PREFIX_0o       1   	/* 0o10 = 8, 0O100 = 64 */
#define PREFIX_0b       1       /* 0b10 = 2, 0B100 = 4 */

#define	CAPS_ON		0	/* do you like (faked) caps lock by default? */
#define	LOWER_CASE_ON	1	/* allow "dup" to find "DUP" by default */
#define LWRCASE_FN_ON	1	/* convert file names to lower case? */
#define FLOAT_INPUT_ON	1	/* allow input of floating point numbers */
#define	NO_COMPLETION	0	/* 1 if you don't like completion feature */


/*
 * Options for the text and block editor:
 */

#ifndef EDITOR
#define	EDITOR		"emacs"	/* preferred editor for text files, */
#endif				/* env-variable EDITOR overrides this */
#ifndef LOGNAME
#define	LOGNAME		"you"	/* if your machine has no env-variable */
#endif				/* LOGNAME, put your name here */
#define	ED_TABW		4	/* width of tab steps in block editor */


/*
 * Special keys, effective when using termunix terminal driver:
 */

#define	INTR_KEY	'\025'	/* (^U) key to send a SIGINT */
#define	QUIT_KEY	'\034'	/* (^\) key to send a SIGQUIT */
#define	SUSP_KEY	'\032'	/* (^Z) key to send a SIGTSTP */

/*
 *
 */
#define SMART_WORDS 	1
 
#ifndef P4_MP
#define P4_MP 1 /* how many machines by default */
#endif
 
#ifndef P4_MP_MAX
#define P4_MP_MAX 8 /* how many machines as maximum */
#endif

#endif 
  /* ndef __P4_OPTIONS_H */
/*@}*/
