/** @name p4utils.c -- utility functinos
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 1.4 %
 *    (%date_modified: Thu Jan 27 21:38:51 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: p4utils.c~1.4:csrc:bln_12xx!1 %1 %1 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include <stdlib.h>
#include <setjmp.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "support.h"
#include "p4_log.h"

void
p4_exit (int n)
{
  longjmp (PFE.loop, 'X'); /* exit mainloop */
}

void
p4_delay (int ms)
/*
 * Somehow wait ms milli-seconds.
 * Versions using poll and select according to Stevens'
 * "Advanced Programming in the UNIX Environment" p.705
 */
{
#if !defined VxWorks
#if defined HAVE_DELAY
  delay (ms);
#elif defined EMX
  _sleep2 (ms);
#elif defined HAVE_USLEEP
  usleep (ms * 1000);
#elif defined HAVE_POLL
  static struct pollfd dummy = { 0, POLLHUP, POLLHUP };
  poll (&dummy, 1, ms);
#elif defined HAVE_SELECT || defined HAVE_SYS_SELECT_H
  struct timeval tval;
  tval.tv_sec = ms / 1000;
  tval.tv_usec = ms % 1000 * 1000;
  select (0, NULL, NULL, NULL, &tval);
#else
  sleep ((ms + 999) / 1000);
#endif
#else /* VxWorks */
# ifndef NANOSLEEP_BUG
  struct timespec rqtp;

  rqtp.tv_sec = (time_t) (ms / 1000);
  rqtp.tv_nsec = (long) (ms % 1000) * 1000000; 
  nanosleep (&rqtp, 0);
# else
  int clocks = CLOCKS_PER_SEC*ms/1000;
  if (clocks) {
      taskDelay (clocks); 
  }
# endif
#endif /* VxWorks */
}

void *
p4_xcalloc (int n_elem, size_t size) /* allocate memory, die when failed */
{
  void *p = calloc (n_elem, size);

  P4_debug3 (13, "xcalloc 0x%p[%i*%i]", p, n_elem, size);

  if (p == NULL)
    p4_fatal ("out of memory");
  
  return p;
}

void *
p4_calloc (int n_elem, size_t size) /* allocate memory, with debug info */
{
  void *p = calloc (n_elem, size);

  if (p)
  {
     P4_debug3 (13, "calloc 0x%p[%i*%i]", p, n_elem, size);
  }else{
     P4_warn2 ("calloc is null[%i*%i]", n_elem, size);
  }

  return p;
}

void *
p4_xalloc (size_t size)	/* allocate memory, throw when failed */
{
  void *p = calloc (1, size);

  P4_debug2 (13, "xalloc 0x%p[%i]", p, size);

  if (p == NULL)
    p4_throw (THROW_OUT_OF_MEMORY);

  return p;
}

void
p4_xfree (void* p)
{
   P4_debug1 (13, "xfree 0x%p", p);
   free (p);
}

/*@}*/

