/** @name p4words.h --- types, codes and macros for symbol table load
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 4.9 %
 *    (%date_modified: Mon May 15 18:00:58 2000 %)
 *
 *  @description
 *         here are the macros that are used to build the various
 *         symbols tables, including wordsets and loadlists.
 */
/*@{*/
#ifndef __P4WORDS_H
#define __P4WORDS_H

#ifndef __P4E_H
#include "p4e.h"
#endif

#ifndef __P4_MACRO_H
#include "p4_macro.h"
#endif

#ifndef __P4_CONST_H
#include "p4_const.h"
#endif

#define P4WORDS(SET)  P4CAT3 (p4_,SET,_Words)
#define P4WLIST(SET)  P4CAT3 (p4_,SET,_WList)

#define extern_P4WORDS(SET) extern const p4Words P4WORDS(SET)

typedef struct			/* describe a word for registration */
{				/* in the dictionary */
  const char *name;		/* name of word */
  p4code ptr;			/* code it executes or something */
}				/* according to spc@pineal.math.fau.edu: */
p4Word;				/* 'void *' doesn't work here on SGI */

typedef struct			/* describes the set of words */
{				/* a module loads into the dictionary */
  int n;			/* how many words? */
  const p4Word *w;		/* pointer to vector of words */
  const char *name;		/* name of word set, or null */
}
p4Words;

#define P4_LISTWORDS( SET )  \
   static const p4Word P4WLIST(SET)[]

# define P4_COUNTWORDS(SET,NAME) \
  const p4Words P4WORDS(SET) = \
  {				\
    DIM (P4WLIST (SET)),	\
    P4WLIST (SET),		\
    NAME			\
  }
  
#ifdef _P4_SOURCE

#define LISTWORDS( SET )  \
   static const p4Word P4WLIST(SET)[]

# define COUNTWORDS(SET,NAME) \
  const p4Words P4WORDS(SET) = \
  {				\
    DIM (P4WLIST (SET)),	\
    P4WLIST (SET),		\
    NAME			\
  }
  
#endif  

/* Encoding the kind of definition i.e. which runtime to fill into the cfa. */
/* Octal 1XY, where X&4 means "immediate" */

#define _CO 0101	/* code definition */
#define _CI 0141	/* immediate code definition */
#define _CS 0142	/* immediate state smart words */

#define _CX 0103	/* init code definition */
#define _CY 0104     	/* deinit code definition */

#define _OV 0105	/* ordinary variable */
#define _OC 0106	/* ordinary constant */
#define _OL 0107	/* ordinary value */

#define _OT 0111        /* ordinary offset */
#define _VO 0112	/* vocabulary */

#define _IV 0145	/* immediate variable */
#define _IC 0146	/* immediate constant */
#define _IL 0147	/* immediate value */

#define _SV 0113	/* system variable */
#define _SC 0114	/* system constant */

#define _DV 0115	/* dictionary related system variable */
#define _DC 0116	/* dictionary related system constant */

/* macros to build entries in the wordlists: */

#define	CO(NM,PCODE)	{ "\101"NM, &P4CODE (PCODE) }
#define	CI(NM,PCODE)	{ "\141"NM, &P4CODE (PCODE) }
#define CS(NM,SEM)	{ "\142"NM, (p4code)&P4SEMANTICS(SEM) }

#define CX(NM,PCODE)	{ "\103"NM, &P4CODE (PCODE) } /* AUTO-INIT */
#define CY(NM,PCODE)    { "\104"NM, &P4CODE (PCODE) } /* AUTO-DEINIT */

#define OV(NM)		{ "\105"NM, (p4code)0) }
#define	OC(NM,VAL)	{ "\106"NM, (p4code)(VAL) }
#define OL(NM,VAL)	{ "\107"NM, (p4code)(VAL) }
#define OT(NM,VAL)      { "\111"NM, (p4code)(VAL) }

#define IV(NM)		{ "\145"NM, (p4code)0) }
#define	IC(NM,VAL)	{ "\146"NM, (p4code)(VAL) }
#define IL(NM,VAL)	{ "\147"NM, (p4code)(VAL) }

/* #define	VO(NM,LIST)	{ "\112"NM, (p4code)(&P4WORDS(LIST,WList)) } */

#define	SV(NM,VAR)	{ "\113"NM, (p4code)&(VAR) }
#define	SC(NM,VAR)	{ "\114"NM, (p4code)&(VAR) }

#if __IBMC__ && __OS2__	/* IBM C Set/2 thinks OFFSET_OF is no constant expr */
#define	DV(NM,MEMBER)	{ "\115"NM, (p4code)do_##MEMBER }
#define	DC(NM,MEMBER)	{ "\116"NM, (p4code)do_##MEMBER }
#else
#define	DV(NM,MEMBER)	{ "\115"NM, (p4code)OFFSET_OF(struct _p4_thread, MEMBER) }
#define	DC(NM,MEMBER)	{ "\116"NM, (p4code)OFFSET_OF(struct _p4_thread, MEMBER) }
#endif


/* ------------------------
   defining a load list
   ------------------------  */
   
#define P4_LOAD_END  ((void*)0)   
#define P4_LOAD_INTO ((void*)1)
#define P4_LOAD_EXEC ((void*)2)
#define P4_LOAD_ORDER ((void*)3)
#define P4_LOAD_SLOT ((void*)4)
#define P4_LOAD_MAX  ((void*)8)
   
#define P4_LOADSTRUCT(NAME) P4CAT3(p4_,NAME,_LoadList)
#define P4_LOADLIST(NAME) const void* P4_LOADSTRUCT(NAME) []
   
#ifdef MODULE        
#define P4_MODULE_LIST(name) \
        void* p4_module_llist() \
        {               	\
           return &P4_LOADSTRUCT(name); \
        }
#else
#define P4_MODULE_LIST(name) \
        void* name##_p4_module_llist() \
        {               	\
           return &P4_LOADSTRUCT(name); \
        }
#endif

#endif 
   /* __P4_WORDS_H */
/*@}*/

