/** @name support.c ---  Subroutines for the Internal Forth-System
 * 
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE            @(#) %derived_by: guidod %
 *  @version %version: 4.58 %
 *    (%date_modified: Thu Apr 27 14:13:58 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: support.c~4.58:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>		/* access() if available */
#endif

#if defined HAVE_PWD_H
#include <pwd.h>		/* to resolve ~user/path */
#endif

#include "dblsub.h"
#include "term.h"
#include "lined.h"
#include "help.h"
#include "missing.h"
#include "support.h"

#include "p4_log.h"

#ifndef isascii 
#define isascii(X) ((unsigned char)(X) < 0x80)
#endif

p4cell
p4_aligned (p4cell n)		/* return cell-aligned address */
{
  while (!P4_ALIGNED (n))
    n++;
  return n;
}

p4cell
p4_dfaligned (p4cell n)	/* return double float-aligned address */
{
  while (!P4_DFALIGNED (n))
    n++;
  return n;
}


/************************************************************************/
/* miscellaneous execution semantics and runtimes                       */
/************************************************************************/

FCode (p4_sysvar_RT)
{
  *--SP = WP_PFA[0];
}

FCode (p4_sysconst_RT)
{
  *--SP = *(p4cell *) WP_PFA[0];
}

FCode (p4_dictvar_RT)
{
  *--SP = (p4cell) ((char *) p4TH + WP_PFA[0]);
}

FCode (p4_dictconst_RT)
{
  *--SP = *(p4cell *) ((char *) p4TH + WP_PFA[0]);
}

void				/* compiles the execution semantics */
p4_compile1 (void)			/* of a state smart word */
{
  FX (p4_Q_comp);
#ifndef USE_FIG
#   if P4_LOG
    if (! ((SemantP4**)p4WP)[-1] ) P4_fail ("<no seman!>");
    if (! ((SemantP4**)p4WP)[-1]->exec[0] ) P4_warn ("<no seman1!>");
#   endif
  
    COMMA (&((SemantP4 **) p4WP)[-1]->exec[0]); 
#else
    /* well, there's nothing good with this solution... *FIXME*/
    {   SemantP4** s = PFE_FIG;
        for(; *s; s++)
        {
           if ((**s).comp == *p4IP[-1])
           {   COMMA(&((**s).exec[0])); 
               return; 
           }
        }
        P4_fail ("<no semant!>");
    }
#endif
}

void				/* compiles the alternative exec.sem. */
p4_compile2 (void)			/* of an even smarter word (e.g. TO) */
{
  FX (p4_Q_comp);
#ifndef USE_FIG
#   if P4_LOG
    if (! ((Seman2P4**)p4WP)[-1] ) P4_fail ("<no seman!>");
    if (! ((Seman2P4**)p4WP)[-1]->exec[1] ) P4_warn ("<no seman2!>");
#   endif

    COMMA (&((Seman2P4 **) p4WP)[-1]->exec[1]);
#else
    /* well, there's nothing good with this solution... *FIXME*/
    {   SemantP4** s = PFE_FIG;
        for(; *s; s++)
        {
           if ((**s).comp == *p4IP[-1])
           {   COMMA(&((**s).exec[1])); 
               return; 
           }
        }
        P4_fail ("<no semant!>");
    }
#endif
}

/************************************************************************/
/* strings                                                              */
/************************************************************************/

void
p4_strpush (const char *s)
{
  if (s)
    *--SP = (p4cell)s, *--SP = strlen (s);
  else
    *--SP = 0, *--SP = 0;
}

char *
p4_pocket (void)
{
  char *p = PFE.pockets[PFE.pocket];

  PFE.pocket = (PFE.pocket + 1) % P4_opt.pockets;
  return p;
}

int
p4_dash_trailing (char *s, int n)
{
  while (n > 0 && isspace ((unsigned char) s[n - 1]))
  {
    n--;
  }
  return n;
}

#ifndef p4_strncmpi 
int
p4_strncmpi (const char* p, const char* q, int n)
#if defined HAVE_STRNCASECMP
{
  return strncasecmp (p, q, n);
}
#elif defined HAVE_STRNICMP
{
  return strnicmp (p, q, n);
}
#else
{
  for(; n; --n )
    {
      if( !*p || !*q ) return p-q;
      if( toupper(*p) != toupper(*q) )
	return n; /* returns the differing tails, like bcmp */
      p++;
      q++;
    }
  return 0;
}
#endif /*HAVE_...*/
#endif /*strncmpi*/

void
p4_lower (char *p, int n)	/* tolower() applied to counted string */
{
  while (--n >= 0)
    *p = tolower (*p), p++;
}

void
p4_upper (char *p, int n)	/* toupper() applied to counted string */
{
  while (--n >= 0)
    *p = toupper (*p), p++;
}

char *
p4_store_c_string (const char *src, int n, char *dst, int max)
{
  if (n >= max)
    n = max - 1;
  memcpy (dst, src, n);
  dst[n] = '\0';
  return dst;
}

char *
p4_store_filename (const char *src, int n, char *dst, int max)
{
  if (n >= max)
    n = max - 1;
  memcpy (dst, src, n);
  dst[n] = '\0';
  if (LOWER_CASE_FN)
    p4_lower (dst, n);
  return dst;
}

/************************************************************************/
/* string comparision and pattern matching                              */
/************************************************************************/

char *
p4_search (const char *p1, int u1, const char *p2, int u2)
/* search for substring p2/u2 in string p1/u1 */
{
  if (u2 == 0)
    return (char *) p1;
  if (u2 > u1)
    return NULL;
  u1 -= u2;
  for (;;)
    {
      char *p = (char *) memchr (p1, *p2, u1 + 1);

      if (p == NULL)
	return NULL;
      if (memcmp (p, p2, u2) == 0)
	return (char *) p;
      u1 -= p - p1;
      if (u1 == 0)
	return NULL;
      p1 = p + 1;
      u1--;
    }
}

static int
do_match (const short *pattern, const char *string, int ic)
/* match with a processed pattern, i.e. one without `\' escapes */
{
  int c;

  for (;;)
    switch (c = *pattern++)
      {
      case '\0':
	return *string == '\0';
      case -'*':
	while (*string && !do_match (pattern, string, ic))
	  string++;
	continue;
      case -'?':
	if (*string++)
	  continue;
	return 0;
      default:
        if (!ic) {
           if (*string++ == c)
              continue;
        } else {
           if (tolower(*string++) == tolower(c))
              continue;
        }
	return 0;
      }
}

int
p4_match (const char *pattern, const char *string, int ic)
/*
 * Match string against pattern.
 * Pattern knows wildcards `*' and `?' and `\' to escape a wildcard.
 */
{
  short buf[0x100], *p = buf;

  /* preprocess pattern, remove `\' */
  for (;;)
    {
      int c = *(unsigned char *) pattern;

      pattern++;
      switch (c)
	{
	default:
	  *p++ = c;
	  continue;
	case '\0':
	  *p = 0;
	  break;
	case '?':
	  *p++ = -'?';
	  continue;
	case '*':
	  *p++ = -'*';
	  continue;
	case '\\':
	  if (*pattern)
	    *p++ = *pattern++;
	  else
	    *p++ = c;
	  continue;
	}
      break;
    }
  /* match with preprocessed pattern */
  return do_match (buf, string, ic);
}

/************************************************************************/
/* expanding file names with paths and extensions                       */
/************************************************************************/

static const char *
get_token (const char *toks, char del, char *s)
/*
 * Sort of strtok() which is useless here since it isn't reentrant and
 * can't handle empty tokens.
 * Isolates a string delimited by del from toks, copies it to s.
 * Returns a pointer to just after the delimiter inside toks
 * or to the delimiting '\0' of toks if the token was the last one.
 */
{
  while (*toks != del)
    {
      if (*toks == '\0')
	{
	  *s = '\0';
	  return toks;
	}
      *s++ = *toks++;
    }
  toks++;
  *s = '\0';
  return toks;
}

static int
try_extensions (char *nm, const char *ext)
/*
 * Append all extensions from ext to nm.
 * Check if file exists, if so return true, else false.
 */
{
  char *z = nm + strlen (nm);

  while (*ext)
    {
      ext = get_token (ext, PATH_DELIMITER, z);
      if (access (nm, F_OK) == 0)
	return 1;
    }
  return 0;
}

static const char *
get_home_dir (const char *path, char *home)
/*
 * Called when the first character of path is a '~', tries to find
 * the home directory. Writes the home directory including a trailing
 * '/' into home and returns the rest of path.
 */
{
  char user[80];
  path = get_token (path + 1, DIR_DELIMITER, user);
  if (user[0] == '\0')
    {
      char *HOME = getenv ("HOME");
      sprintf (home, "%s%c", HOME ? HOME : "~", DIR_DELIMITER);
    }
  else
    {
#if HAVE_PWD_H
      struct passwd *passwd = getpwnam (user);
      if (passwd)
	sprintf (home, "%s%c", passwd->pw_dir, DIR_DELIMITER);
      else
#endif
	sprintf (home, "~%s%c", user, DIR_DELIMITER);
    }
  return path;
}

char *
p4_expand_filename (const char *nm, const char *paths,
		 const char *exts, char *fn)
/*
 * nm    file name input, short
 * paths search path for files
 * ext   default file extensions
 * fn    full file name, output
 */
{
  char path[PATH_LENGTH];

  if (*nm == DIR_DELIMITER)
    {
      strcpy (path, nm);
      if (try_extensions (path, exts))
	return strcpy (fn, path);
    }
  else if (*nm == '~')
    {
      strcat (path, get_home_dir (nm, path));
      if (try_extensions (path, exts))
	return strcpy (fn, path);
    }
  else
    {
      while (*paths)
	{
	  char *p;
	  paths = get_token (paths, PATH_DELIMITER, path);
	  if (*path == '~')
	    {
	      char buf [PATH_LENGTH];
	      strcpy (buf, path);
	      strcat (path, get_home_dir (buf, path));
	    }
	  p = path + strlen (path);
	  if (p > path && p[-1] != DIR_DELIMITER)
	    *p++ = DIR_DELIMITER;
	  strcpy (p, nm);
	  if (try_extensions (path, exts))
	    return strcpy (fn, path);
	}
    }
  return strcpy (fn, nm);
}

/************************************************************************/
/* unsigned and floored divide and number i/o conversion                */
/************************************************************************/

udiv_t
p4_udiv (p4ucell num, p4ucell denom)	/* unsigned divide procedure, single prec */
{
  udiv_t res;

  res.quot = num / denom;
  res.rem = num % denom;
  return res;
}

fdiv_t
p4_fdiv (p4cell num, p4cell denom)	/* floored divide procedure, single prec */
{
  fdiv_t res;

  res.quot = num / denom;
  res.rem = num % denom;
  if (res.rem && (num ^ denom) < 0)
    {
      res.quot--;
      res.rem += denom;
    }
  return res;
}

p4ucell
p4_u_d_div (p4udcell *ud, p4ucell denom)
/*
 * Divides *ud by denom, leaves result in *ud, returns remainder.
 * For number output conversion: dividing by BASE.
 */
{
  p4udcell nom = *ud;
  udiv_t h;

  h = p4_udiv (P4xD0 (nom), denom);
  P4xD0 (*ud) = h.quot;
  P4xD0 (nom) = h.rem;
  h = p4_udiv (nom.hi, denom);
  P4xD1 (*ud) = h.quot;
  P4xD1 (nom) = h.rem;
  h = p4_udiv (P4xCELL (P4xD1 (nom), P4xD2 (nom)), denom);
  P4xD2 (*ud) = h.quot;
  P4xD2 (nom) = h.rem;
  h = p4_udiv (nom.lo, denom);
  P4xD3 (*ud) = h.quot;
  return h.rem;
}

void
p4_u_d_mul (p4udcell *ud, p4ucell w, p4ucell c)
/*
 * Computes *ud * w + c, where w is actually only half of a cell in size.
 * Leaves result in *ud.
 * For number input conversion: multiply by BASE and add digit.
 */
{
  c += P4xD3 (*ud) * w, P4xD3 (*ud) = P4xW1 (c), c >>= HALFCELL;
  c += P4xD2 (*ud) * w, P4xD2 (*ud) = P4xW1 (c), c >>= HALFCELL;
  c += P4xD1 (*ud) * w, P4xD1 (*ud) = P4xW1 (c), c >>= HALFCELL;
  P4xD0 (*ud) = P4xD0 (*ud) * w + c;
}

int
p4_dig2num (p4char c, p4ucell *n, p4ucell base)
/*
 * Get value of digit c into *n, return flag: valid digit.
 */
{
  if (c < '0')
    return P4_FALSE;
  if (c <= '9')
    c -= '0';
  else
    {
      if (LOWER_CASE)
	c = toupper (c);
      if (c < 'A')
	return P4_FALSE;
      if (c <= 'Z')
	c -= 'A' - ('9' - '0' + 1);
      else
	{
	  if (LOWER_CASE || c < 'a')
	    return P4_FALSE;
	  c -= 'a' - ('9' - '0' + 1) - ('Z' - 'A' + 1);
	}
    }
  if (c >= base)
    return P4_FALSE;
  *n = c;
  return P4_TRUE;
}

char
p4_num2dig (p4ucell n)		/* make digit */
{
  if (n < 10)
    return n + '0';
  if (n < 10 + 'Z' - 'A' + 1)
    return n - 10 + 'A';
  else
    return n - (10 + 'Z' - 'A' + 1) + 'a';
}

void
p4_hold (char c)			/* insert into pictured numeric */
{				/* output string */
  if (P4HLD <= (char *) DP)
    p4_throw (THROW_PICNUM_OVER);
  *--P4HLD = c;
}

const char *
p4_to_number (const char *p, p4ucell *n, p4udcell *d, p4ucell base)
{
  for (; *n > 0; p++, --*n)
    {
      p4ucell c;

      if (!p4_dig2num (*p, &c, base))
	break;
      p4_u_d_mul (d, base, c);
      if (P4DPL >= 0)
	P4DPL++;
    }
  return p;
}

int
p4_number_question (const char *p, p4ucell n, p4dcell *d)
{
  p4ucell base = 0;
  int sign = 0;
  
  for (; n; p++, n--)
    {
      switch (*p)
	{
	default:
	  break;
	case '-':
	  if (sign)
	    return 0;
	  sign = 1;
	  continue;
#if PREFIX_HEX
	case PREFIX_HEX:
	  if (base)
	    return 0;
	  base = 16;
	  continue;
#endif
#if PREFIX_BINARY
	case PREFIX_BINARY:
	  if (base)
	    return 0;
	  base = 2;
	  continue;
#endif
	}
      break;
    }

#if PREFIX_0x || PREFIX_0o || PREFIX_0b
  if( ! base && n > 2 && *p == '0' ) 
    { 
      switch(*(p+1))
      {
#     if (PREFIX_0x)
          case 'x':
          case 'X':
	      base = 16; p+=2; n-=2;
	      break;
#     endif
#     if (PREFIX_0o)
          case 'o':
          case 'O':
	      base = 8; p+=2; n-=2;
	      break;
#     endif
#     if (PREFIX_0b)
          case 'b':
          case 'B':
	      base = 2; p+=2; n-=2;
	      break;
#     endif
      }
    }
#endif
  if (base == 0)
      base = BASE;
  d->lo = d->hi = 0;
  P4DPL = -1;
  p = p4_to_number (p, &n, (p4udcell *) d, base);
  if (n == 0)
    goto happy;
  if (*p != '.') 	
    return 0;   	
  P4DPL = 0;
  p++;
  n--;
  p = p4_to_number (p, &n, (p4udcell *) d, base);
  if (n != 0) 		
    return 0; 		
happy:
  if (sign)
    dnegate (d);
  return 1;
}

#ifndef P4_NO_FP
#if defined USE_STRTOD		/* most systems have good strtod */

static int
p4_to_float (char *p, p4cell n, double *r)
{
  char buf[80], *q;
  
  if (!*p) return 0; /* strtod does crash on vxworks being empty non-null *gud*/

  p4_store_c_string (p, n, buf, sizeof buf);
  if (tolower (buf[n - 1]) == 'e')
      buf[n++] = '0';
  buf[n] = '\0';
  *r = strtod (buf, &q);
  if (q == NULL)
    return 1;
  while (isspace (*q))
    q++;
  return *q == '\0';
}

#else				/* but some haven't */

static int
p4_to_float (char *p, p4cell n, double *r)
{
  enum state 			/* states of the state machine */
  {
    bpn,			/* before point, maybe sign */
    bp,				/* before point, no more sign (had one) */
    ap,				/* after point */
    exn,			/* exponent, maybe sign */
    ex,				/* exponent, no more sign, yet no digit */
    exd				/* exponent, no more sign, had one digit */
  };
  enum state state = bpn;
  int sign = 1;			/* sign of mantissa */
  long double mant = 0;		/* the mantissa */
  int esign = 1;		/* sign of exponent */
  int exp = 0;			/* the exponent */
  int bdigs = 0;		/* digits before point */
  int scale = 0;		/* number of digits after point */

  while (--n >= 0)
    {
      char c = *p++;

      switch (state)
	{
	case bpn:
	  switch (c)
	    {
	    case '-':
	      sign = -1;
	    case '+':
	      state = bp;
	      continue;
	    case '.':
	      state = ap;
	      continue;
	    default:
	      if (isspace (c))
		continue;
	      if (isdigit (c))
		{
		  bdigs = 1;
		  mant = c - '0';
		  state = bp;
		  continue;
		}
	    }
	  return 0;
	case bp:
	  switch (c)
	    {
	    case '.':
	      state = ap;
	      continue;
#if 0
	    case '-':
	      esign = -1;
	    case '+':
	      state = ex;
	      continue;
#endif
	    case 'e':
	    case 'E':
	      state = exn;
	      continue;
	    default:
	      if (isdigit (c))
		{
		  bdigs++;
		  mant *= 10;
		  mant += c - '0';
		  continue;
		}
	    }
	  return 0;
	case ap:
	  switch (c)
	    {
	    case '-':
	      esign = -1;
	    case '+':
	      state = ex;
	      continue;
	    case 'e':
	    case 'E':
	      state = exn;
	      continue;
	    default:
	      if (isdigit (c))
		{
		  mant *= 10;
		  mant += c - '0';
		  scale--;
		  continue;
		}
	    }
	  return 0;
	case exn:
	  state = ex;
	  switch (c)
	    {
	    case '-':
	      esign = -1;
	    case '+':
	      continue;
	    default: ;
	    }
	case ex:
	  if (isdigit (c))
	    {
	      exp = c - '0';
	      state = exd;
	      continue;
	    }
	  return 0;
	case exd:
	  if (isdigit (c))
	    {
	      exp *= 10;
	      exp += c - '0';
	      continue;
	    }
	  return 0;
	}
    }
  *r = sign * mant * pow10 (scale + esign * exp);
  return bdigs - scale > 0;
}

#endif
#endif /* _NO_FP */
/*
 * These are for internal use only (SEE and debugger),
 * The real `UD.R' etc. words use HOLD and the memory area below PAD
 */

char *
p4_str_ud_dot_r (p4udcell ud, char *p, int w, int base)
{
  *--p = '\0';
  do
    {
      *--p = p4_num2dig (p4_u_d_div (&ud, base));
      w--;
    }
  while (ud.lo || ud.hi);
  while (w > 0)
    *--p = ' ', w--;
  return p;
}

char *
p4_str_d_dot_r (p4dcell d, char *p, int w, int base)
{
  int sign = 0;

  if (d.hi < 0)
    dnegate (&d), sign = 1;
  *--p = '\0';
  do
    {
      *--p = p4_num2dig (p4_u_d_div ((p4udcell *) &d, base));
      w--;
    }
  while (d.lo || d.hi);
  if (sign)
    *--p = '-', w--;
  while (w > 0)
    *--p = ' ', w--;
  return p;
}

char *
p4_str_dot (p4cell n, char *p, int base)
{
  p4dcell d;
  char *bl;

  *--p = '\0';
  bl = p - 1;
  d.lo = n;
  d.hi = n < 0 ? -1 : 0;
  p = p4_str_d_dot_r (d, p, 0, base);
  *bl = ' ';
  return p;
}

/************************************************************************/
/* console i/o                                                          */
/************************************************************************/

/* output adjusting the OUT variable */

void
p4_outc (char c)			/* emit single character */
{
  int x, y;

  p4_putc (c);
  p4_wherexy (&x, &y);
  OUT = x;
}

void
p4_outs (const char *s)		/* type a string */
{
  int x = 0, y = 0;

  p4_puts (s);
  p4_wherexy (&x, &y);
  OUT = x;
}

int
p4_outf (const char *s,...)	/* type a string with formatting */
{
  char buf[0x200];
  va_list p;
  int r;

  va_start (p, s);
  r = vsprintf (buf, s, p);
  p4_outs (buf);
  va_end (p);
  return r;
}

void
p4_type (const char *s, p4cell n)	/* TYPE counted string to terminal */
{
  int x, y;

  while (--n >= 0)
    p4_putc_noflush (*s++);
  p4_wherexy (&x, &y);
  OUT = x;
  p4_put_flush ();
}

void
p4_type_on_line (const char *s, p4cell n)
{
  if (OUT + n >= PFE.cols)
    FX (p4_cr);
  p4_type (s, n);
}

void
p4_spaces (int n)
{
  int x, y;

  while (--n >= 0)
    p4_putc_noflush (' ');
  fflush (stdout);
  p4_wherexy (&x, &y);
  OUT = x;
}

void
p4_tab (int n)
{
  p4_spaces (n - OUT % n);
}

void
p4_dot_line (File *fid, p4cell n, p4cell l)
{
  char *p = p4_block (fid, n) + l * 64;
  p4_type (p, p4_dash_trailing (p, 64));
}

/* input */

static int
p4_get_line (char *p, p4cell n)
{
  char *q, buf[0x100];
  extern FCode (p4_bye);

  q = fgets (buf, n, stdin);
  if (q == NULL)
    FX (p4_bye);
  q = strrchr (q, '\n');
  if (q)
    *q = '\0';
  strcpy (p, buf);
  return strlen (p);
}

int
p4_expect_canonical (char *p, p4cell n)	
                      	/* EXPECT counted string from terminal, */
{			/* no editing, no echo */
  int i;		
  char c;
  int out = 0;

  for (i = 0; i < n;)
    {
      switch (c = p4_getkey ())
	{
	default:
	  p[i++] = c; out++;
	  continue;
	case '\t':
	  while (i < n)
	    {
	      p[i++] = ' '; out++;
	      if (out % 8 == 0)
		break;
	    }
	  continue;
        case '\33':
	case '\r':
	case '\n':
	  goto fin;
	case 127:
	case '\b':
	  if (i <= 0)
	      continue;
	  i--; out--;
	  continue;
	}
    }
fin:
  p[i] = 0;
  SPAN = i;
  return i;
}

int
p4_expect (char *p, p4cell n)	/* EXPECT counted string from terminal, */
{				/* simple editing facility with Backspace, */
  int i;			/* very traditional, use lined() instead! */
  char c;

  if (P4_opt.canonical == 2)
    return p4_expect_canonical (p, n);
  if (P4_opt.canonical)
    return p4_get_line (p, n);
  for (i = 0; i < n;)
    {
      switch (c = p4_getkey ())
	{
	default:
	  p[i++] = c;
	  p4_outc (c);
	  continue;
	case 27:
	  for (; i > 0; i--)
	    FX (p4_backspace);
	  continue;
	case '\t':
	  while (i < n)
	    {
	      p[i++] = ' ';
	      FX (p4_space);
	      if (OUT % 8 == 0)
		break;
	    }
	  continue;
	case '\r':
	case '\n':
	  FX (p4_space);
	  goto fin;
	case 127:
	case '\b':
	  if (i <= 0)
	    {
	      p4_dot_bell ();
	      continue;
	    }
	  i--;
	  FX (p4_backspace);
	  continue;
	}
    }
fin:p[i] = 0;
  SPAN = i;
  return i;
}

int
p4_accept (char *p, int n)		/* better input facility using lined() */
{
  if (P4_opt.canonical == 2)
    return p4_expect_canonical (p, n);
  if (P4_opt.canonical)
    return p4_get_line (p, n);
  PFE.accept_lined.string = p;
  PFE.accept_lined.max_length = n;
  p4_lined (&PFE.accept_lined, NULL);
  FX (p4_space);
  return PFE.accept_lined.length;
}

int
p4_Q_stop (void)		/* check for 'q' pressed */
{
  if (p4_ekeypressed ())
  {
      register int ch;  	
      ch = p4_getkey ();  
      if (tolower (ch) == 'q') return 1;
  }
  return 0;
}

int
p4_Q_cr (void)
/*
 * Like CR but stop after one screenful and return flag if 'q' pressed.
 * Improved by aph@oclc.org (Andrew Houghton)
 */
{
  static char more[] = "more? ";
  static char help[] = "\r[next line=<return>, next page=<space>, quit=q] ";

  FX (p4_cr);
  if (P4_opt.canonical)
    return 0;
  if (PFE.lines < PFE.more)
    return 0;
  PFE.lines = 0;
  for (;;)
    {
      register int ch;
      p4_outs (more);
      ch = p4_getkey (); 	/* tolower(..) may be a macro ! *gud*/
      switch (tolower (ch)) 
	{
	case 'n':		/* no more */
	case 'q':		/* quit    */
	  return 1;
	case 'y':		/* more    */
	case ' ':		/* page    */
	  while (OUT)
	    FX (p4_backspace);
	  PFE.more = PFE.rows - 1;
	  return 0;
	case '\r':		/* line    */
	case '\n':		/* line    */
	  while (OUT)
	    FX (p4_backspace);
	  PFE.more = 1;
	  return 0;
	default:		/* unknown */
	  p4_dot_bell ();
	  /* ... */
	case '?':		/* help    */
	case 'h':		/* help    */
	  p4_outs (help);
	  break;
	}
    }
}

/** a little helper to check return codes */
int 
p4_err (int err)
{
  if (err >= 0) return err;
  P4_warn1 ("error: %s", strerror (-err));
  return err;
}


/************************************************************************/
/* files                                                                */
/************************************************************************/

void
p4_file_errorz (const char *fn)
{
  p4_throw (-1024 - errno, fn);
}

void
p4_file_error (const char *fn, int len)
{
  char buf[PATH_LENGTH];

  p4_store_filename (fn, len, buf, sizeof buf);
  p4_file_errorz (buf);
}

static File *
p4_free_file_slot (void)
{
  File *f;

  for (f = PFE.files; f < PFE.files_top; f++)
    if (f->f == NULL)
      {
	memset (f, 0, sizeof *f);
	return f;
      }
  return NULL;
}

int
p4_file_access (const char *fn, int len)
/*
 * Return best possible access method,
 * 0 if no access but file exists, -1 if file doesn't exist.
 */
{
  char buf[PATH_LENGTH];

  p4_store_filename (fn, len, buf, sizeof buf);
  if (access (buf, F_OK) != 0)
    return -1;
  if (access (buf, R_OK | W_OK) == 0)
    return FMODE_RW;
  if (access (buf, R_OK) == 0)
    return FMODE_RO;
  if (access (buf, W_OK) == 0)
    return FMODE_WO;
  return 0;
}

static char open_mode[][4] =	/* mode strings for fopen() */
{
  "r", "r+", "r+",		/* R/O W/O R/W */
  "rb", "r+b", "r+b",		/* after application of BIN */
};

File *
p4_open_file (const char *name, int len, int mode)
{
  File *fid;

  fid = p4_free_file_slot ();
  if (fid == NULL)
    return NULL;
  p4_store_filename (name, len, fid->name, sizeof fid->name);
  fid->mode = mode;
  fid->last_op = 0;
  strcpy (fid->mdstr, open_mode[mode - FMODE_RO]);
  if ((fid->f = fopen (fid->name, fid->mdstr)) == NULL)
    return NULL;
  fid->size = (p4ucell) (fsize (fid->f) / BPBUF);
  fid->n = (unsigned) -1;
  return fid;
}

File *
p4_create_file (const char *name, int len, int mode)
{
  char fn[PATH_LENGTH];
  File *fid;

  p4_store_filename (name, len, fn, sizeof fn);
  fclose (fopen (fn, "wb"));
  fid = p4_open_file (name, len, mode);
  if (fid)
    {
      return fid;
    }
  else
    {
      remove (fn);
      return NULL;
    }
}

int
p4_close_file (File *fid)
{
  int res = 0;

  if (fid->f)
    {
      res = fclose (fid->f);
      memset (fid, 0, sizeof *fid);
    }
  return res;
}

int
p4_reposition_file (File *fid, long pos)
{
  fid->last_op = 0;
  return fseek (fid->f, pos, SEEK_SET) ? errno : 0;
}

static int
p4_can_read (File *fid)
/*
 * Called before trying to read from a file.
 * Checks if you may, maybe fseeks() so you can.
 */
{
  switch (fid->mode)		/* check permission */
    {
    case FMODE_WO:
    case FMODE_WOB:
      return 0;
    }
  if (fid->last_op < 0)		/* last operation was write? */
    fseek (fid->f, 0, SEEK_CUR); /* then seek to this position */
  fid->last_op = 1;
  return 1;
}

static int
p4_can_write (File *fid)
/*
 * Called before trying to write to a file.
 * Checks if you may, maybe fseeks() so you can.
 */
{
  switch (fid->mode)		/* check permission */
    {
    case FMODE_RO:
    case FMODE_ROB:
      return 0;
    }
  if (fid->last_op > 0)		/* last operation was read? */
    fseek (fid->f, 0, SEEK_CUR); /* then seek to this position */
  fid->last_op = -1;
  return 1;
}

int
p4_read_file (void *p, p4ucell *n, File *fid)
{
  int m;

  if (!p4_can_read (fid))
    return EPERM;
  errno = 0;
  m = fread (p, 1, *n, fid->f);
  if (m != (int) *n)
    {
      *n = m;
      return errno;
    }
  else
    return 0;
}

int
p4_write_file (void *p, p4ucell n, File *fid)
{
  if (!p4_can_write (fid))
    return EPERM;
  errno = 0;
  return (p4ucell) fwrite (p, 1, n, fid->f) != n ? errno : 0;
}

int
p4_resize_file (File *fid, long size)
{
  long pos;
  int r;

  if (fid == NULL || fid->f == NULL)
    p4_throw (THROW_FILE_NEX);

  pos = ftell (fid->f);
  if (pos == -1)
    return -1;

  fclose (fid->f);
  r = fn_resize (fid->name, size);
  fid->f = fopen (fid->name, fid->mdstr);

  if (pos < size)
    fseek (fid->f, pos, SEEK_SET);
  else
    fseek (fid->f, 0, SEEK_END);
  return r;
}

int
p4_read_line (char *p, p4ucell *u, File *fid, p4cell *ior)
{
  int c, n;

  if (!p4_can_read (fid))
    return EPERM;
  fid->pos = ftell (fid->f);
  for (n = 0; (p4ucell) n < *u; n++)
    switch (c = getc (fid->f))
      {
      default:
	*p++ = c;
	continue;
      case EOF:
	*u = n;
	if (ferror (fid->f))
	  *ior = errno;
	else
	  *ior = 0;
	return P4_FLAG (n > 0);
      case '\r':
	c = getc (fid->f);
	if (c != '\n')
	  ungetc (c, fid->f);
      case '\n':
	goto happy;
      }
happy:
  *u = n;
  *ior = 0;
  fid->n++;
  return P4_TRUE;
}

#ifndef NO_SYSTEM
int
p4_systemf (const char *s,...)	/* issue a system() call after formatting */
{
  char buf[0x100];
  va_list p;
  int r;

  va_start (p, s);
  vsprintf (buf, s, p);
  va_end (p);
  p4_system_terminal ();
  p4_swap_signals ();
  r = system (buf);
  p4_swap_signals ();
  p4_interactive_terminal ();
  p4_dot_normal ();
  return r;
}
#endif /* NO_SYSTEM */

/************************************************************************/
/* source input								*/
/************************************************************************/

/* 1. read from terminal */

FCode (p4_query)
{
  SOURCE_ID = 0;
  BLK = 0;
  TO_IN = 0;
  TIB = PFE.tib;
  NUMBER_TIB = p4_accept (TIB, TIB_SIZE);
  SPAN = NUMBER_TIB;
}

/* 2. read from text-file */

int
p4_next_line (void)
{
  p4cell ior;
  p4ucell len;

  len = sizeof SOURCE_FILE->buffer;
  if (!p4_read_line (SOURCE_FILE->buffer, &len, SOURCE_FILE, &ior))
    {
      SOURCE_FILE->len = len;
      return 0;
    }
  TIB = SOURCE_FILE->buffer;
  NUMBER_TIB = SOURCE_FILE->len = len;
  BLK = 0;
  TO_IN = 0;
  return 1;
}

/* 3. read from block-file */

File *
p4_open_block_file (const char *name, int len)
{
  char nm[PATH_LENGTH], fn[PATH_LENGTH];
  int mode;

  p4_store_filename (name, len, nm, sizeof nm);
  p4_expand_filename (nm, P4_opt.blkpaths, P4_opt.blkext, fn);
  mode = p4_file_access (fn, strlen (fn));
  if (mode <= 0)
    return NULL;
  return p4_open_file (fn, strlen (fn), mode + FMODE_BIN);
}

int
p4_use_block_file (const char *name, int len)
{
  File *fid;

  fid = p4_open_block_file (name, len);
  if (fid == NULL)
    return P4_FALSE;
  if (BLOCK_FILE)
    {
      FX (p4_save_buffers);
      p4_close_file (BLOCK_FILE);
    }
  BLOCK_FILE = fid;
  return P4_TRUE;
}

void
p4_read_write (File *fid, char *p, p4ucell n, int readflag)
/* very traditional block read/write primitive */
{
  size_t len;

  p4_Q_file_open (fid);
  clearerr (fid->f);
  if (n > fid->size)
    p4_throw (THROW_INVALID_BLOCK);
  if (readflag && n == fid->size)
    {
      memset (p, ' ', BPBUF);
      return;
    }
  if (fseek (fid->f, n * BPBUF, SEEK_SET) != 0)
    p4_file_errorz (fid->name);
  if (readflag)
    {
      if (!p4_can_read (fid))
	p4_throw (THROW_BLOCK_READ);
      len = fread (p, 1, BPBUF, fid->f);
      if (ferror (fid->f))
	p4_file_errorz (fid->name);
      memset (p + len, ' ', BPBUF - len);
    }
  else
    {
      if (!p4_can_write (fid))
	p4_throw (THROW_BLOCK_WRITE);
      len = fwrite (p, 1, BPBUF, fid->f);
      if (len < BPBUF || ferror (fid->f))
	p4_file_errorz (fid->name);
      if (n == fid->size)
	fid->size++;
    }
  return;
}

char *
p4_buffer (File *fid, p4ucell n, int *reload)
{
  p4_Q_file_open (fid);
  if (fid->n != n)
    {
      if (fid->updated)
	p4_read_write (fid, fid->buffer, fid->n, P4_FALSE);
      fid->n = n;
      *reload = 1;
    }
  else
    *reload = 0;
  return fid->buffer;
}

char *
p4_block (File *fid, p4ucell n)
{
  char *p;
  int reload;

  p = p4_buffer (fid, n, &reload);
  if (reload)
    p4_read_write (fid, p, n, P4_TRUE);
  return p;
}

void
p4_empty_buffers (File *fid)
{
  p4_Q_file_open (fid);
  ZERO (fid->buffer);
  fid->n = UINT_MAX;
  fid->updated = 0;
}

void
p4_save_buffers (File *fid)
{
  if (fid && fid->updated)
    {
      p4_read_write (fid, fid->buffer, fid->n, P4_FALSE);
      fflush (fid->f);
      fid->updated = 0;
    }
}

void
p4_update (File *fid)
{
  p4_Q_file_open (fid);
  if ((int) fid->n < 0)
    p4_throw (THROW_INVALID_BLOCK);
  fid->updated = 1;
}

void
p4_list (File *fid, int n)
{
  int i;

  for (i = 0; i < 16; i++)
    {
      FX (p4_cr);
      p4_outf ("%2d: ", i);
      p4_dot_line (fid, n, i);
    }
  FX (p4_space);
  SCR = n;
}

/* dispatch input source */

void
p4_source (char **p, int *n)
{
  switch (SOURCE_ID)
    {
    case -1:			/* string from EVALUATE */
      *p = TIB;			
      *n = NUMBER_TIB;
      break;
    case 0:			/* string from QUERY or BLOCK */
      if (BLK)
	{
	  *p = p4_block (BLOCK_FILE, BLK);
	  *n = BPBUF;
	}
      else
	{                       
	  *p = TIB;
	  *n = NUMBER_TIB;
	}
      break;
    default:			/* source line from text file */
      *p = SOURCE_FILE->buffer; 
      *n = SOURCE_FILE->len;
    }
}

void *
p4_save_input (void *p)
{
  Iframe *iframe = (Iframe *) p;

  --iframe;
  iframe->magic = INPUT_MAGIC;
  iframe->input = PFE.input;
  iframe->prev = PFE.saved_input;
  PFE.saved_input = iframe;

  return ((void*) iframe);
}

void *
p4_restore_input (void *p)
{
  Iframe *iframe = (Iframe *) p;

  if (iframe->magic != INPUT_MAGIC)
    p4_throw (THROW_ARG_TYPE);
  PFE.input = iframe->input;
  PFE.saved_input = iframe->prev;
  ++iframe;

  p = (void *) iframe;
  return p;
}

int
p4_refill (void)
{
  switch (SOURCE_ID)
    {
    case -1:
      return 0;
    case 0:
      if (BLK)
	{
	  BLK++;
	  TO_IN = 0;
	}
      else
	FX (p4_query);
      return 1;
    default:
      return p4_next_line ();
    }
}

void
p4_skip_delimiter (char del)
{
  char *q;
  int i, n;

  p4_source (&q, &n);
  if (del == ' ')
  {
    for (i = TO_IN;
         i < n && isascii (q[i]) && isspace ((unsigned char) q[i]);
         i++)
    {
      ;
    }
  }
  else
  {
    for (i = TO_IN; i < n && q[i] == del; i++)
    {
      ;
    }
  }
  TO_IN = i;
}

int
p4_parse (char del, char **p, p4ucell *l)
{
  char *q;
  int i, n; 

  p4_source (&q, &n);
  *p = q + TO_IN;
  if (del == ' ')
  {
    for (i = TO_IN;
         i < n && !(isascii (q[i]) && isspace ((unsigned char) q[i]));
         i++)
    {
      ;
    }
  }
  else
  {
    for (i = TO_IN; i < n && q[i] != del; i++)
    {
      ;
    }
  }
  *l = i - TO_IN;
  if (i == n)
  {
    TO_IN = i;
    return 0;
  }
  else
  {
    TO_IN = i + 1;
    return 1;
  }
}

char *
p4_word (char del)			/* action of WORD callable from C functions */
{
  char *p, *q;
  int n, i;

  p4_skip_delimiter (del);
  p4_source (&q, &n);
  q += TO_IN;
  n -= TO_IN;
  p = (char *) DP + 1;
  if (del == ' ')
  {
    for (i = 0; i < n && !(isascii (*q) && isspace ((unsigned char) *q)); i++)
    {
      *p++ = *q++;
    }
  }
  else
  {
    for (i = 0; i < n && *q != del; i++)
    {
      *p++ = *q++;
    }
  }
  TO_IN += i + (i < n);
  *p = '\0';
  if (i > 255)
  {
    p4_throw (THROW_PARSE_OVER);
  }
  *DP = i;
  return (char *) DP;
}

/*gud*
  return and args mean the same as for ~parse, 
  but it really scans like ~word. It most cases you can replace
  word(.) with a sequence of p4_parseword (.) and p4_hereword (.);
  The point is, that p4_parseword _doesn't_ copy the next word onto
  here, it just returns the pointers. In some cases, esp. where
  a failure could be p4_thrown , it must be copied to here later.
  You can use p4_hereword (.) for that. See interpret() for an example.
*/
int
p4_parseword (char del, char** p, p4ucell* len)
{
  char *q;
  int n, i;
  
  p4_skip_delimiter (del);
  p4_source (&q, &n);	
  q += TO_IN;
  n -= TO_IN;
  *p = q;
  if (del == ' ')
  {
    for (i = 0; i < n && !(isascii (*q) && isspace ((unsigned char) *q)); i++)
    {
      q++;
    }
  }
  else
  {
    for (i = 0; i < n && *q != del; i++)
    {
      q++;
    }
  }
  *len = i;
  TO_IN += i + (i<n);

  return (i<n);
}

char*
p4_hereword (char* p, p4ucell len)
{
  if (len > 255)
    p4_throw (THROW_PARSE_OVER);
  memcpy (DP+1, p, len);
  *DP = len;
  return DP;
}


/************************************************************************/
/* inner and outer interpreter						*/
/************************************************************************/

typedef struct			/* jumpbuf is a jmp_buf enhanced by */
{				/* space to preserve global variables */
  jmp_buf jmp;			/* stored in registers. */
#ifdef P4_REGRP
  p4xt **rp;
#endif
#ifdef P4_REGSP
  p4cell *sp;
#endif
#ifdef P4_REGLP
  p4cell *lp;
#endif
#ifdef P4_REGFP
  double *fp;
#endif
}
jumpbuf;

FCode (p4_jump)			/* longjmp via (jumpbuf*) following inline */
{				/* purpose: stop the inner interpreter */
  jumpbuf *buf = (jumpbuf *) *IP;

#ifdef P4_REGRP		/* save global register variables */
  buf->rp = RP;
#endif
#ifdef P4_REGSP
  buf->sp = SP;
#endif
#ifdef P4_REGLP
  buf->lp = LP;
#endif
#ifndef P4_NO_FP
#ifdef P4_REGFP
  buf->fp = FP;
#endif
#endif
  longjmp (buf->jmp, 1);
}

void
p4_run_forth (p4xt xt)		/* Run a forth word from within C-code. */
{				/* This is the inner interpreter. */
  static p4code jump_p = PFX (p4_jump);
  jumpbuf stop;
  p4xt list[3];

  list[0] = xt;
  list[1] = &jump_p;
  list[2] = (p4xt) &stop;
  IP = list;
#ifndef P4_WP_VIA_IP
  p4WP = *IP;
#endif

  if (setjmp (stop.jmp))
    {
#ifdef P4_REGRP		/* restore global register variables */
      RP = stop.rp;		/* clobbered by longjmp() */
#endif
#ifdef P4_REGSP
      SP = stop.sp;
#endif
#ifdef P4_REGLP
      LP = stop.lp;
#endif
#ifndef P4_NO_FP
#ifdef P4_REGFP
      FP = stop.fp;
#endif
#endif
      return;
    }
  /* next_loop */
  for (;;)
    {
#ifdef P4_WP_VIA_IP
      p4xt w;
#define NEXT w = *IP++, (*w) ()	/* ip is register but p4WP isn't */
#else
#define NEXT p4WP = *IP++, (*p4WP) ()	
				/* ip and p4WP are same: register or not */
#endif

#ifdef UNROLL_NEXT		/* if it helps */
      NEXT; NEXT; NEXT; NEXT;	/* do a little loop unrolling for speed */
      NEXT; NEXT; NEXT; NEXT;
#else
      NEXT;			/* on some machines it doesn't do any good */
#endif
    }
}

void
p4_call (p4xt xt)
{
#if 0 && defined SYS_sunos
  void *saved_ip;

  saved_ip = IP;
  printf ("%X/%X\n", IP, saved_ip);
  p4_run_forth (xt);
  printf ("%X/%X\n", IP, saved_ip);
  IP = saved_ip;
  printf ("%X/%X\n\n", IP, saved_ip);
#else
  p4xt *saved_ip = IP;

  p4_run_forth (xt);
  IP = saved_ip;

#endif
}

void
p4_normal_execute (p4xt xt)
{
#ifdef P4_WP_VIA_IP
  p4_call (xt);
#else
  p4WP = xt;
  (*p4WP) ();
#endif
}

FCode (p4_interpret)
{
  char *s, *p;
  int len;
  p4dcell d;
#ifndef P4_NO_FP
  double f;
#endif

  for (;;)
  {
      for (;;)
      {
	  p4_parseword (' ', &p, &len); /* does'nt copy to HERE !! */
	  if (len) break;

	  switch (SOURCE_ID)
          {
	  default:
	      if (p4_next_line ())
		continue;
	  case 0:
	  case -1:
	      return;
          }
      }  
      if (STATE)
      {
	  if (PFE.locals && p4_compile_local (p, len))
	    continue;
	  s = p4_find (p, len);
	  if (s != NULL)
	  {
	      p4xt xt = p4_name_from (s);
	      if (*_FFA(s) & P4xIMMEDIATE)
              {   p4_hereword (p, len); /* call(xt) could p4_throw ... */
		  p4_call (xt);
		  FX (p4_Q_stack);
		  continue;
              }else
		  COMMA (xt);
	      continue;
	  }
	  if (PFE.smart_char)
          {   p4xt xt = PFE.smart_char (*p);
	      if (xt) 
              {	  TO_IN -= len; p4_skip_delimiter (*p);
                  p4_call (xt);
		  continue;
	      }
	  }

	  if (p4_number_question (p, len, &d))
	  {
	      if (P4DPL >= 0) 
              {   COMPILE1 (p4_two_literal);
		  COMMA (d.hi);
              }else
		  COMPILE1 (p4_literal);
	      COMMA (d.lo);
	      continue;
	  }
#         ifndef P4_NO_FP
  	  if (BASE == 10 && FLOAT_INPUT && p4_to_float (p, len, &f)) 
          {
#             if DFLOAT_ALIGN > CELL_ALIGN
	      if (P4_DFALIGNED (DP))
		COMPILE2 (p4_f_literal);
#             endif
	      COMPILE1 (p4_f_literal);
	      FCOMMA (f);
	      continue;
          } 
#         endif
      }else{ /*(!STATE)*/
	  s = p4_find (p, len);
	  if (s != NULL) 
          {   p4_hereword (p, len); /* call(name_from(s)) could p4_throw ... */
              p4_call (p4_name_from (s));
              FX (p4_Q_stack);

	      continue;
          }
	  if (PFE.smart_char) 
          {   p4xt xt = PFE.smart_char (*p);
	      if (xt) 
              {   TO_IN -= len; p4_skip_delimiter (*p);
                  p4_call (xt);
		  continue;
	      }
	  }
	  if (p4_number_question (p, len, &d)) 
          {   *--SP = d.lo;
	      if (P4DPL >= 0) *--SP = d.hi;
	      continue;
	  }
#         ifndef P4_NO_FP
	  if (BASE == 10 && FLOAT_INPUT && p4_to_float (p, len, &f)) 
          {   *--FP = f;
	      continue;
	  }
#         endif
      }
      p4_hereword (p, len); /* p4_throw show the word at HERE */
      p4_throw (THROW_UNDEFINED);
  }
}

void
p4_evaluate (char *p, int n)
{
  RP = (p4xt **) p4_save_input (RP);
  SOURCE_ID = -1;
  BLK = 0;
  TIB = p;
  NUMBER_TIB = n;
  TO_IN = 0;
  FX (p4_interpret);
  RP = (p4xt **) p4_restore_input (RP);
}

void
p4_load (File *fid, p4ucell blk)
{
  if (blk == 0)
    p4_throw (THROW_INVALID_BLOCK);
  RP = (p4xt **) p4_save_input (RP);
  BLOCK_FILE = fid;
  SOURCE_ID = 0;
  BLK = blk;
  TO_IN = 0;
  FX (p4_interpret);
  RP = (p4xt **) p4_restore_input (RP);
}

void
p4_load_file (char *fn, int cnt, int blk)
{
  File *fid = p4_open_block_file (fn, cnt);

  if (fid == NULL)
     p4_file_error (fn, cnt);
  p4_load (fid, blk);
}

void
p4_thru (File *fid, int lo, int hi)
{
  int i;

  for (i = lo; i <= hi; i++)
    p4_load (fid, i);
}

void
p4_include_file (File *fid)
{
  if (fid == NULL || fid->f == NULL)
      p4_throw (THROW_FILE_NEX);
  RP = (p4xt **) p4_save_input (RP);
  SOURCE_ID = (p4cell) fid;
  BLK = 0;
  TO_IN = 0;
  FX (p4_interpret);
  RP = (p4xt **) p4_restore_input (RP);
}

int
p4_included1 (const char *name, int len, int throws)
{
  char nm[0x80], fn[0x100];
  File *f;

  p4_store_filename (name, len, nm, sizeof nm);
  p4_expand_filename (nm, P4_opt.incpaths, P4_opt.incext, fn);
  f = p4_open_file (fn, strlen (fn), FMODE_RO);
  if (!f)
  {  if (throws) p4_throw (THROW_FILE_NEX); 
     else{ 
          P4_fail2 ("- could not open '%s' (paths='%s')\n", fn, P4_opt.incpaths); 
          return 0; 
     }
  }
  p4_include_file (f);
  p4_close_file (f);
  return 1;
}

void
p4_included (const char* name, int len)
{
  p4_included1 (name, len, 1);
}

static void
p4_unnest_input (Iframe *p)
{
  while (PFE.saved_input && PFE.saved_input != p)
    {
      switch (SOURCE_ID)
	{
	case -1:
	case 0:
	  break;
	default:
	  p4_close_file (SOURCE_FILE);
	}
      RP = (p4xt **) p4_restore_input (PFE.saved_input);
    }
}

/************************************************************************/
/* QUIT, ABORT and exception handling					*/
/************************************************************************/

int
p4_mainloop_cpu (int n);

FCode (p4_ok)
{
  if (!STATE)
  {
    p4_outs ("ok");
    if (P4_opt.cpu_nr) {
       p4_outc ('-');
       p4_outc ('0'+PFE.nr); /*fixmee - makes MP_MAX bound to 10 */
    }
    FX (p4_space);
  }
}

static void
quit_initializations (void)	/* Things quit has to initialize: */
{
  RP = PFE.r0;			/* return stack */
  LP = NULL;			/* including all local variables */
  STATE = P4_FALSE;		/* interpreting now */
  PFE.cAtch = NULL;		/* and no exceptions caught */
  p4_debug_off ();		/* turn off debugger */
}

static void
abort_initializations (void)	/* Things ABORT has to initialize: */
{
  SP = PFE.s0;			/* stacks */
#ifndef P4_NO_FP
  FP = PFE.f0; 
#endif
  FX (p4_reset_order);		/* reset search order */
  FX (p4_definitions);		/* and vocabulary in extension */
  FX (p4_decimal);			/* number i/o base */
  FX (p4_standard_io);		/* disable i/o redirection */
}

 /* single cpu code, which should give you the outline of a p4th-mainloop */
 /* in some modules, we actually save the cpu's PFE.loop, 
    and put our own mainloop in there (so abort and quit will return there)
 */
int
p4_mainloop ()
{
   register int err;
   switch (err=setjmp (PFE.loop))
   {
   case  0:  /* newloop -> do abort*/
      /* initialize */
   case 'A': /* do abort */
      abort_initializations ();
      /* -> do quit */
   case 'Q': /* do quit */
      quit_initializations ();
   case 'S': /* schedule, mp only */
				/* normal interactive QUIT */
                                /* doing the QUERY-INTERPRET loop */
      p4_unnest_input (NULL);
      for (;;)
      {
      	FX (p4_cr);		
        FX (p4_query);		
        FX (p4_interpret);		
        FX (p4_Q_stack);	
        FX (p4_ok);
      }

   case 'X': /* exit / bye */
      /* cleanup (); */
      return 0;
   }
   return err;
}

/* and here comes the Multi-Forth Mainloop */

int
p4_mainloop_cpu (int n)
{
  if (n >= 0 && n <= P4_opt.cpus && P4_opt.cpu[n])
     return (P4_opt.cpu_nr = n);
  return P4_opt.cpu_nr;
}

void
p4_goto_cpu (int n)
{
  p4_mainloop_cpu (n);
  longjmp (PFE.loop, 'S'); /* schedule now */
}

/*
 *  important: scheduling to a different cpu means to reassign p4TH,
 *             and that may only occur in such a mainloop.
 */

int
p4_mainloop_mp () /*FIXME* didn't check for multi-register reassignments */
{
   register int err;

   p4TH = P4_opt.cpu[0];   
   P4_opt.cpu_nr = 0;
   {
     next:
       if (P4_opt.cpu_nr >= P4_opt.cpus) 
          p4_goto_cpu (0); /* inited all cpus, so jump *forward* */

       if (!P4_opt.cpu[P4_opt.cpu_nr])
          p4_goto_cpu (0); /* run beyond, hopefully cpu[0] is valid */
       p4TH = P4_opt.cpu[P4_opt.cpu_nr];

       switch (err=setjmp (PFE.loop))
       {
       case 0:
          abort_initializations ();
          quit_initializations ();
          P4_opt.cpu_nr++;
          goto next;

          /* reinitialize */
       case 'A': /* do abort */
          abort_initializations ();
          /* -> do quit */
       case 'Q': /* do quit */
          quit_initializations ();
       case 'S': /* schedule */
          {
            _schedule:
              if (!P4_opt.cpu[P4_opt.cpu_nr])
                 P4_opt.cpu_nr = 0; /* run beyond */
              p4TH = P4_opt.cpu[P4_opt.cpu_nr];
				        /* normal interactive QUIT */
                                        /* doing the QUERY-INTERPRET loop */
              p4_unnest_input (NULL);
              for (;;)
              {
                FX (p4_ok);
      	        FX (p4_cr);
                FX (p4_query);
                FX (p4_interpret);
                FX (p4_Q_stack);
                if (PFE.nr != P4_opt.cpu_nr)
                    goto _schedule; /* schedule implicitly */
              }
          }
       case 'X': /* exit / bye */
          /* cleanup (); */
          return 0;
       }
       return err;
   }
}

static void
show_error (char *fmt,...)
{
  char buf[256];
  va_list p;
  int n;

  PFE.input_err = PFE.input;	/* save input specification of error */
  va_start (p, fmt);
  vsprintf (buf, fmt, p);
  va_end (p);
  p4_outf ("\nError: \"%.*s\" %s\n", *DP, DP + 1, buf);
  switch (SOURCE_ID)
    {
    case 0:
      if (BLK && BLOCK_FILE && !ferror (BLOCK_FILE->f))
	{
          p4_outf ("Block %lu line %d:\n",
		(unsigned long) BLK, (int) TO_IN / 64);
          p4_dot_line (BLOCK_FILE, BLK, TO_IN / 64);
	  n = TO_IN % 64;
	  break;
	}
    case -1:
      p4_type (TIB, NUMBER_TIB);
      n = TO_IN;
      break;
    default:
      p4_outf ("File %s line %lu:\n",
	    SOURCE_FILE->name, (unsigned long) SOURCE_FILE->n);
      p4_type (TIB, NUMBER_TIB);
      n = TO_IN;
    }
  p4_outf ("\n%*s", n, "^");
  longjmp (PFE.loop, 'A'); /*abort*/
}

static void
throw_msg (int id, char *msg)
{
  /* *INDENT-OFF* */
  static char *throw_explanation[] =
    {
      /*  -1 */ NULL, /* ABORT */
      /*  -2 */ NULL, /* ABORT" */
      /*  -3 */ "stack overflow",
      /*  -4 */ "stack underflow",
      /*  -5 */ "return stack overflow",
      /*  -6 */ "return stack underflow",
      /*  -7 */ "do-loops nested too deeply during execution",
      /*  -8 */ "dictionary overflow",
      /*  -9 */ "invalid memory address",
      /* -10 */ "division by zero",
      /* -11 */ "result out of range",
      /* -12 */ "argument type mismatch",
      /* -13 */ "undefined word",
      /* -14 */ "interpreting a compile-only word",
      /* -15 */ "invalid FORGET",
      /* -16 */ "attempt to use a zero-length string as a name",
      /* -17 */ "pictured numeric output string overflow",
      /* -18 */ "parsed string overflow",
      /* -19 */ "definition name too long",
      /* -20 */ "write to a read-only location",
      /* -21 */ "unsupported operation",
      /* -22 */ "control structure mismatch",
      /* -23 */ "address alignment exception",
      /* -24 */ "invalid numeric argument",
      /* -25 */ "return stack imbalance",
      /* -26 */ "loop parameters unavailable",
      /* -27 */ "invalid recursion",
      /* -28 */ "user interrupt",
      /* -29 */ "compiler nesting",
      /* -30 */ "obsolescent feature",
      /* -31 */ ">BODY used on non-CREATEDd definition",
      /* -32 */ "invalid name argument",
      /* -33 */ "block read exception",
      /* -34 */ "block write exception",
      /* -35 */ "invalid block number",
      /* -36 */ "invalid file position",
      /* -37 */ "file I/O exception",
      /* -38 */ "non-existent file",
      /* -39 */ "unexpected end of file",
      /* -40 */ "invalid BASE for floating-point conversion",
      /* -41 */ "loss of precision",
      /* -42 */ "floating-point divide by zero",
      /* -43 */ "floating-point result out of range",
      /* -44 */ "floating-point stack overflow",
      /* -45 */ "floating-point stack underflow",
      /* -46 */ "floating-point invalid argument",
      /* -47 */ "compilation word list deleted",
      /* -48 */ "invalid POSTPONE",
      /* -49 */ "search-order overflow",
      /* -50 */ "search-order underflow",
      /* -51 */ "compilation word list changed",
      /* -52 */ "control flow stack overflow",
      /* -53 */ "exception stack overflow",
      /* -54 */ "floating-point underflow",
      /* -55 */ "floating-point unidentified fault",
      /* -56 */ NULL, /* QUIT */
      /* -57 */ "error in sending or receiving a character",
      /* -58 */ "[IF], [ELSE] or [THEN] error",
      /* -59 */ "dictionary space exhausted"
    },
  *throw_explanation_p4[] =
    {
      /* -2048 */ "no or not matching binary image",
      /* -2049 */ "binary image too big",
      /* -2050 */ "out of memory",
      /* -2051 */ "index out of range",
    };
  /* *INDENT-ON* */

  if (-1 - DIM (throw_explanation) < id && id <= -1)
    {
      /* ANS-Forth throw codes, messages are in throw_explanation[] */
      strcpy (msg, throw_explanation[-1 - id]);
    }
  else if (-1024 < id && id <= -256)
    {
      /* Signals, see signal.c, those not handled and not fatal lead
         to THROW */
      char sigmsg [0x80];
      strcpy (sigmsg, msg);
      sprintf (msg, "Received signal %d, %s", -256 - id, sigmsg);
    }
  else if (-2048 < id && id <= -1024)
    {
      /* File errors */
      char fn[PATH_LENGTH];
      strcpy (fn, msg);
      sprintf (msg, ": File %s: %s", fn, strerror (-1024 - id));
    }
  else if (-2048 - DIM (throw_explanation_p4) < id && id <= -2048)
    {
      strcpy (msg, throw_explanation_p4[-2048 - id]);
    }
  else
    {
      sprintf (msg, "%d THROW unassigned", id);
    }
}


int
p4_catch (p4xt xt)
{
  Except *x = P4_DEC (RP, Except);
  int id;

  x->magic = EXCEPTION_MAGIC;
  x->ipp = IP;
  x->spp = SP;
  x->lpp = LP;
#ifndef P4_NO_FP
  x->fpp = FP;
#endif
  x->iframe = PFE.saved_input;
  x->prev = PFE.cAtch;
  PFE.cAtch = x;
  id = setjmp (x->jmp);
  if (!id)
     p4_call (xt);
  PFE.cAtch = x->prev;
  RP = (p4xt **) &x[1];
  return id;
}

void
p4_throw (int id,...)
{
  Except *x = PFE.cAtch;
  va_list p;
  char msg[256];
  
#ifdef _K12_SOURCE
  extern void trcStack(int); /* show stack trace */
  extern int taskIdSelf ();
  extern int taskPriorityGet (int, int*);
  extern int taskDelay (int);
  extern int taskSpawn (char*, int, int, int, void (*)(int), int);
  int taskid, taskprio;
  if (p4_LogMask & P4_LOG_DEBUG) 
  { /* if any debug-channel used */
    taskPriorityGet ((taskid= taskIdSelf()), &taskprio);
    taskSpawn (0, taskprio, 0, 8192, trcStack, taskid);
    taskDelay (1); /* sched_yield */
  }
#endif
  
  if (PFE.throw_cleanup) 
    { PFE.throw_cleanup ();
      PFE.throw_cleanup = NULL;
    }

  if (x && x->magic == EXCEPTION_MAGIC)
    {
      IP = x->ipp;
      SP = x->spp;
      LP = x->lpp;
#ifndef P4_NO_FP
      FP = x->fpp;
#endif P4_NO_FP
      p4_unnest_input (x->iframe);
      longjmp (x->jmp, id);
    }
  switch (id)
    {
    case THROW_ABORT_QUOTE:
      {
	char *addr;
	int len;

	va_start (p, id);
	addr = va_arg (p, char *);
	len = va_arg (p, int);
	va_end (p);
	show_error ("%.*s", len, addr);
      }
    case THROW_ABORT:
      longjmp (PFE.loop, 'A'); /*abort*/
    case THROW_QUIT:
      longjmp (PFE.loop, 'Q'); /*quit*/
    default:
      if (-2048 < id && id <= -256)
	{
	  va_start (p, id);
	  strcpy (msg, va_arg (p, char *));
	  va_end (p);
	}
      throw_msg (id, msg);
      show_error (msg);
    }
}

void
p4_abortq (const char *fmt,...)
{
  char buf[128];
  int n;
  va_list p;

  va_start (p, fmt);
  n = vsprintf (buf, fmt, p);
  va_end (p);
  p4_throw (THROW_ABORT_QUOTE, buf, n);
}

void
p4_Q_pairs (p4cell n)
{
  if (n != *SP++)
     p4_throw (THROW_CONTROL_MISMATCH);
}

void
p4_Q_file_open (File *fid)
{
  if (fid == NULL || fid->f == NULL)
     p4_throw (THROW_FILE_NEX);
}

/************************************************************************/
/* Initialize dictionary, and system variables, include files		*/
/************************************************************************/

void				/* set up all system variables */
p4_initialize_system (void)	/* and initialize the dictionary */
{
  /* memset (p4TH, 0, sizeof (*p4TH)); now done earlier *gud*/
  if (PFE.nr) { printf (" CPU%i ", PFE.nr); }
  
  SP = PFE.s0;
#ifndef P4_NO_FP
  FP = PFE.f0;
#endif
  RP = PFE.r0;
  TIB = PFE.tib;
  BASE = 10;
  P4DPL = -1;
  PRECISION = 6;
  WORDL_FLAG = P4_opt.lower_case_on; /* implicitly enables HASHing */
  LOWER_CASE_FN = P4_opt.lower_case_fn;
  FLOAT_INPUT = P4_opt.float_input;
  PFE.local = (char (*)[32]) PFE.stack;

  memset (PFE.files_top - 3, 0, sizeof (File) * 3);

  PFE.stdIn = PFE.files_top - 3;
  PFE.stdIn->f = stdin;
  strcpy (PFE.stdIn->name, "<STDIN>");
  strcpy (PFE.stdIn->mdstr, "r");
  PFE.stdIn->mode = FMODE_RO;

  PFE.stdOut = PFE.files_top - 2;
  PFE.stdOut->f = stdout;
  strcpy (PFE.stdOut->name, "<STDOUT>");
  strcpy (PFE.stdOut->mdstr, "a");
  PFE.stdOut->mode = FMODE_WO;

  PFE.stdErr = PFE.files_top - 1;
  PFE.stdErr->f = stderr;
  strcpy (PFE.stdErr->name, "<STDERR>");
  strcpy (PFE.stdErr->mdstr, "a");
  PFE.stdErr->mode = FMODE_WO;

  if (P4_opt.block_file)
    {
      if (!p4_use_block_file (P4_opt.block_file, strlen (P4_opt.block_file))
	  && strcmp (P4_opt.block_file, DEFAULT_BLKFILE) != 0)
	p4_fatal ("Can't find block file %s", P4_opt.block_file);
    }

/*  read_help_index (HELPDIR, "index"); */

  REDEFINED_MSG = P4_FALSE;

  /* Wipe the dictionary: */
  memset (PFE.dict, 0, (PFE.dictlimit - PFE.dict));
  p4_preload_only ();
  FX (p4_only_RT);
  {
    /* Defines the following default search order:
     * FORTH EXTENSIONS ONLY */
#ifndef MODULE0
#define MODULE0 extensions
#endif
    extern P4_LOADLIST (forth);
    extern P4_LOADLIST (MODULE0); 

#ifdef MODULE1
    extern P4_LOADLIST (MODULE1);
#endif
#ifdef MODULE2
    extern P4_LOADLIST (MODULE2);
#endif
#ifdef MODULE3
    extern P4_LOADLIST (MODULE3);
#endif

    p4_load_llist ((void**) &P4_LOADSTRUCT(forth), ONLY, 0);
    p4_load_llist ((void**) &P4_LOADSTRUCT(MODULE0), ONLY, 0);
    
#ifdef MODULE1
    p4_load_llist ((void**) &P4_LOADSTRUCT(MODULE1), ONLY, 0);
#endif
#ifdef MODULE2
    p4_load_llist ((void**) &P4_LOADSTRUCT(MODULE2), ONLY, 0);
#endif
#ifdef MODULE3
    p4_load_llist ((void**) &P4_LOADSTRUCT(MODULE3), ONLY, 0);
#endif
    if (PFE.set->loadlist[0]) p4_load_llist (PFE.set->loadlist[0], ONLY, 0);
    if (PFE.set->loadlist[1]) p4_load_llist (PFE.set->loadlist[1], ONLY, 0);
    if (PFE.set->loadlist[2]) p4_load_llist (PFE.set->loadlist[2], ONLY, 0);
    if (PFE.set->loadlist[3]) p4_load_llist (PFE.set->loadlist[3], ONLY, 0);
  }
  FX (p4_default_order);

  FENCE = DP;
  LAST  = NULL;
  

  /* Action of ABORT and QUIT, but don't enter the interactive QUIT */
  abort_initializations ();
  quit_initializations ();

  /* Include .pferc if it exists: */
  if (P4_opt.p4rc_file)
    {
      char p4rc [PATH_LENGTH];
      if (*P4_opt.p4rc_file == '~')
	strcat (p4rc, get_home_dir (P4_opt.p4rc_file, p4rc));
      else
	strcpy (p4rc, P4_opt.p4rc_file);
      if (access (p4rc, R_OK) == 0)
	p4_included (p4rc, strlen (p4rc));
    }

  /* Include file from command line: */
  if (P4_opt.include_file)
    p4_included1 (P4_opt.include_file, strlen (P4_opt.include_file), 0);

  REDEFINED_MSG = P4_TRUE;
}

/*@}*/

