/** @name sysdep.c --- definitions that may be missing in your system
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE             @(#) %derived_by: guidod %
 *  @version %version: 4.25 %
 *    (%date_modified: Thu Jan 27 21:31:06 2000 %)
 *
 *  @description
 *              put here any definitions missing in your system -
 *		most of this file contributed by Marko Teiste
 * 
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: sysdep.c~4.25:csrc:bln_12xx!4 %4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <assert.h>

#if defined SYS_WATCOM
#include <i86.h>
#endif

#include "support.h"
#include "nonansi.h"
#include "missing.h"

/* ========================================================================= */
#ifndef HAVE_AH_TRIG
#if !defined P4_NO_FP && !defined VxWorks
/*
 * Simple acosh(), asinh(), atanh() for those unfortunates who don't
 * have them. These are oversimplified routines (no error or boundry
 * checking). !!! DONT TRUST THESE ROUTINES !!!
 */
#include <math.h>

double acosh (double n)
	{ return log (n + sqrt (n * n - 1)); }
double asinh (double n)
	{ return (n < 0 ? -1.0 : 1.0) * log (fabs (n) + sqrt (n * n + 1)); }
double atanh (double n)
	{ return log (1.0 + ((2.0 * n) / (1.0 - n))) * 0.5; }

#endif /* _NO_FP */
#endif /*HAVE_AH_TRIG*/

#ifndef HAVE_STRDUP

#include <stdlib.h>
#include <string.h>

char *
strdup (const char *s)
{
  static char failed [] = "\n<calloc failed>\n";
  char *p = (char *) p4_calloc (1, strlen (s) + 1);
  return p ? strcpy (p, s) : failed;
}

#endif /*HAVE_STRDUP*/

#ifndef HAVE_MEMMOVE

void
memmove (char *d, const char *s, unsigned n)
{
  if (n)
    if (s > d)
      do
	*d++ = *s++;
      while (--n > 0);
    else
      do
	{
	  --n;
	  d [n] = s [n];
	}
      while (n > 0);
}

#endif /*HAVE_MEMMOVE*/

#ifndef HAVE_RENAME

int
rename (const char *source, const char *target)
/*
 * Rename file,
 * This is not a foolproof routine, one of those
 * "I'll do it better when I have more time" things. -mte
 */
{
  char save_name [PATH_LENGTH+1]; /* Name of saved file */

  if (access(target, 0) == 0)
    {
      sprintf (save_name, "%s~", target);
      if (access (save_name, 0) == 0 ||
	  unlink (save_name) == -1 ||
	  link (target, save_name) == -1 ||
	  unlink (target) == -1)
	return -1;
    }
  if (link (source, target) == -1)
    {
      if (save_name != NULL)
	if (link (save_name, target) == -1)
	  return -1;
	else
	  unlink (save_name);
      return -1;
    }
  if (unlink (source) == -1)
    return -1;
  if (save_name != NULL)
    unlink (save_name);
  return 0;
}

#endif /*HAVE_RENAME*/

#ifndef HAVE_FCNTL
int
fcntl (int fildes,
       int cmd,
       ... /* arg */)
{
  return 0;
}
#endif /*HAVE_FCNTL*/

/*@}*/

