/** @name PFE-TERM  ---   terminal i/o, system independent parts
 * 
 *  Copyright Tektronix Inc., Berlin, 1998
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.41 %
 *    (%date_modified: Thu Apr 27 14:15:15 2000 %)
 *
 *  @description
 *       this wordset exports words to talk to the terminal driver
 *       of forth. Depending on the active terminal driver, not
 *       all of these words are fully functional.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: term.c~4.41:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "term.h"
#include "missing.h"
#include "support.h"

#ifndef isascii
#define isascii(X) ((unsigned char) (X) < 0x80)
#endif

/* map function and cursor keys to wordstar key sequences or to EKEY codes: */

static char			/* replacement strings for the key codes, */
*raw2ws [NO_OF_KEYS] =		/* wordstar-like version for block editor: */
{
  "\013L",			/* F1 -> ^KL i.e. interpret this line */
  "\013D",			/* F2 -> ^KD i.e. stamp screen */
  "\013U",			/* F3 -> ^KU i.e. use other block file */
  NULL,
  "\013X",			/* F5 -> ^KX i.e. push line */
  "\013E",			/* F6 -> ^KE i.e. pop line */
  "\013Y",			/* F7 -> ^KY i.e. push and delete line */
  "\013W",			/* F8 -> ^KW i.e. pop and insert line */
  "\013B",			/* F9 -> ^KB i.e. load this block */
  "\025",			/* F10 -> ^U i.e. quit editor */

  NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL,

  "\023", "\004", "\005", "\030",
  "\021S","\021D","\003", "\022",
  "\010", "\007", "\026", "\026",
  "\016", "\021Y","\031", "\021Z",
},
*raw2ekey [NO_OF_KEYS] =	/* version for EKEY */
{
  "\000", "\001", "\002", "\003", "\004",
  "\005", "\006", "\007", "\010", "\011",

  "\012", "\013", "\014", "\015", "\016",
  "\017", "\020", "\021", "\022", "\023",

  "\024", "\025", "\026", "\027",
  "\030", "\031", "\032", "\033",
  "\034", "\035", "\036", "\037",
  "\040", "\041", "\042", "\043",
};

static int
get_remap_key (p4char *replace[])
{
  int i, n, m;

  if (PFE.keyptr)
  {  if (*PFE.keyptr == '\0')
        PFE.keyptr = NULL;
     else
        return *PFE.keyptr++;
  }
  n = 0;
  for (;;)
  {  m = -1;
     PFE.keybuf[n++] = p4_getkey ();
     if (PFE.rawkey_string)
     {  for (i = m; ++i < NO_OF_KEYS;)
	{  if (!PFE.rawkey_string[i] || !replace[i])
	      continue;
	   if (memcmp (PFE.keybuf, PFE.rawkey_string[i], n) == 0)
	   {  if (PFE.rawkey_string[i][n] == '\0')
              {  PFE.keyptr = replace[i];
		 return *PFE.keyptr++ | 0x100;
              }else{
	         m = i;
              }
           }
	}
     }
     if (m < 0)
     {  PFE.keybuf[n] = '\0';
	PFE.keyptr = PFE.keybuf;
	return *PFE.keyptr++;
     }
  }
}

int
p4_getekey (void)
{
  int key  = get_remap_key ((p4char **) raw2ekey);

  if (key == EKEY_kb)
    return '\b';
  return key;
}

int
p4_getwskey (void)
{
  return get_remap_key ((p4char **) raw2ws) & ~0x100;
}

int
p4_ekeypressed (void)
{
  return (PFE.keyptr && *PFE.keyptr) || p4_keypressed ();
}

int
p4_isprint (int c)
{
  return
#if defined IBMPC_CHARSET
    !isascii (c) || !iscntrl (c)	/* almost anything can be displayed */
#else
    isprint (c)
#if defined (ISO_CHARSET)
    || (0xA0 <= (p4char) c && (p4char) c <= 0xFF)
#endif
#endif
    ;
}

void				/* write character to screen */
p4_putc_printable (int c)	/* ensure visible result */
{
  if (p4_isprint (c))
      p4_putc (c);
  else{
      p4_dot_standout_on ();
      p4_putc ((p4char) c < ' ' ? c + '@' : '?');
      p4_dot_standout_off ();
  }
}

int
p4_change_case (int ch)
{
  if (ch >= 0x100)  return ch;
  if (islower (ch)) return toupper (ch);
  if (isupper (ch)) return tolower (ch);
#if defined ISO_CHARSET
  if (0xC0 <= (p4char)ch && (p4char)ch <= 0xDE) return ch + 0x20;
  if (0xE0 <= (p4char)ch && (p4char)ch <= 0xFE) return ch - 0x20;
#elif defined IBMPC_CHARSET
  {
    static char lower [] =
      "\x81\x82\x83\x84\x85\x86\x87\x88" /* IBM 437 lower case characters */
      "\x89\x8A\x8B\x8C\x8D\x91\x93\x94"
      "\x95\x96\x97\x98\xA0\xA1\xA2\xA3"
      "\xA4";
    static char upper [] =
      "\x9A\x90\x41\x8E\x41\x8F\x80\x45" /* IBM 437 upper case characters */
      "\x45\x45\x49\x49\x49\x92\x4F\x99"
      "\x4F\x55\x55\x59\x41\x49\x4F\x55"
      "\xA5";
    char *p;

    if ((p = strchr (lower, (char)ch))) return upper [p - lower];
    if ((p = strchr (upper, (char)ch))) return lower [p - upper];
  }
#endif
  return ch;
}


/*
 * hooks for window size change and job control:
 */



/*
 *  new terminal driver implementation
 */

int  p4_tty_interrupt_key (char ch) { return (*PFE.term->tty_interrupt_key)(ch); }
void p4_interactive_terminal () { (*PFE.term->interactive_terminal)(); }
void p4_system_terminal ()      { (*PFE.term->system_terminal)(); }
void p4_query_winsize ()        { (*PFE.term->query_winsize)(); }
int  p4_keypressed ()           { return (*PFE.term->c_keypressed)(); }
int  p4_getkey ()		{ return (*PFE.term->c_getkey)(); }

void p4_putc_noflush (char c)   { (*PFE.term->c_putc_noflush)(c); }
void p4_putc (char c)           { (*PFE.term->c_putc)(c); }
void p4_puts (const char* s)    { (*PFE.term->c_puts)(s); }
void p4_put_flush ()            { (*PFE.term->c_put_flush) (); }

void p4_gotoxy(int x, int y)	{ (*PFE.term->c_gotoxy)(x,y); }
void p4_wherexy(int* x, int* y)	{ (*PFE.term->c_wherexy)(x,y); }

void p4_goleft (void)		{ (*PFE.term->tput) (P4_TERM_GOLEFT); }
void p4_goright (void)		{ (*PFE.term->tput) (P4_TERM_GORIGHT); }
void p4_goup (void)		{ (*PFE.term->tput) (P4_TERM_GOUP); }
void p4_godown (void) 		{ (*PFE.term->tput) (P4_TERM_GODOWN); }

void p4_dot_home (void)		{ (*PFE.term->tput) (P4_TERM_HOME); }
void p4_dot_clrscr (void)	{ (*PFE.term->tput) (P4_TERM_CLRSCR); }
void p4_dot_clreol (void)	{ (*PFE.term->tput) (P4_TERM_CLREOL); }
void p4_dot_clrdown (void)	{ (*PFE.term->tput) (P4_TERM_CLRDOWN); }
void p4_dot_bell (void)		{ (*PFE.term->tput) (P4_TERM_BELL); }

void p4_dot_normal (void)      	{ (*PFE.term->tput) (P4_TERM_NORMAL); }
void p4_dot_standout_on (void) 	{ (*PFE.term->tput) (P4_TERM_BOLD_ON); }
void p4_dot_standout_off (void) { (*PFE.term->tput) (P4_TERM_BOLD_OFF); }
void p4_dot_bright (void)      	{ (*PFE.term->tput) (P4_TERM_BRIGHT); }
void p4_dot_reverse (void)    	{ (*PFE.term->tput) (P4_TERM_REVERSE); }
void p4_dot_blinking (void)    	{ (*PFE.term->tput) (P4_TERM_BLINKING); }
void p4_dot_underline_on (void) { (*PFE.term->tput) (P4_TERM_UNDERLINE_ON); }
void p4_dot_underline_off (void){ (*PFE.term->tput) (P4_TERM_UNDERLINE_OFF); }

static char tckeycode[][3] =
{
  "k1", "k2", "k3", "k4", "k5",	/* keys in same order as enum keycode */
  "k6", "k7", "k8", "k9", "k0",	/* from term.h */
  "F1", "F2", "F3", "F4", "F5",
  "F6", "F7", "F8", "F9", "FA",
  "kl", "kr", "ku", "kd",	/* result is just what has to be exported */
  "kh", "kH", "kN", "kP",	/* via variable rawkey_string */
  "kb", "kD", "kM", "kI",
  "kA", "kE", "kL", "kC"
};

static char tcctlcode[][3] =
{
  "cm", "ho",
  "le", "nd", "up", "do",
  "cl", "cd", "ce", "bl",
  "dc", "dl",
  "sf", "sr",
  "so", "se", "us", "ue",
  "md", "mr", "mb", "me",
  "ks", "ke"
};

/** SHOW-TERM-CONTROLS ( -- ) for debugging
 * show the current mappings for the terminal output 
 * may give hints about what is wrong if the output
 * seems to miss some functionality
 */
FCode (p4_show_control_strings)
{
  if (PFE.term)
     p4_outf ("\n term control '%s'", PFE.term->name);
  else
     p4_outs ("\n term control unknown");

   if (PFE.control_string) {
      /* for your information why screen manipulation doesn't work :-) */
      int i;
      char *p;
    
      for (i = 0; i < DIM (tcctlcode); i++)
      {
          p4_outf ("\n\"%s\"=", tcctlcode[i]);
          if ((p = PFE.control_string[i]))
            while (*p)
              p4_putc_printable (*p++);
          else
            p4_puts ("undefined");
      }
   } else {
      p4_outs ("\n no controls set. ");
   }
}

/** SHOW-TERM-ESC-KEYS ( -- ) for debugging
 * show the current mappings for the terminal input
 * may give hints about what is wrong if the input
 * seems to miss some functionality
 */
FCode (p4_show_rawkey_strings)
{
   if (PFE.term)
      p4_outf ("\n term rawkeys '%s'", PFE.term->name);
   else
      p4_outs ("\n term rawkeys unknown");

   if (PFE.rawkey_string) {
      /* for your information why function keys don't work :-) */
      int i;
      char *p;
    
      for (i = 0; i < DIM (tckeycode); i++)
        {
          p4_outf ("\n\"%s\"=", tckeycode[i]);
          if ((p = PFE.rawkey_string[i]))
            while (*p)
              p4_putc_printable (*p++);
          else
            p4_puts ("undefined");
        }
   } else {
      p4_outs ("\n no rawkeys set. ");
   }
}

/** SHOW-TERMCAP ( -- ) for debugging
 * print the termcap strings used for input and output
 * may give hints about what is wrong if the terminal
 * seems to miss some functionality
 */
FCode (p4_show_termcap)
{
  FX (p4_show_control_strings);
  FX (p4_show_rawkey_strings);
}

char *p4_dumbterm_controls[] =	/* Some hardcoded vt100 sequences. */
{
  "",			/* cm - cursor move */
  "\f",			/* ho - home position */

  "\b",			/* le - cursor left */
  " ",			/* nd - right one column */
  "",			/* up - up one column */
  "\n",			/* do - down one column */

  "",			/* cl - clear screen and home */
  "",			/* cd - clear down */
  "",			/* ce - clear to end of line */
  "\a",			/* bl - bell */

  "",			/* dc - delete character in line */
  "",			/* dl - delete line from screen */

  "",			/* sf - scroll screen up */
  "",			/* sr - scroll screen down */

  "",			/* so - enter standout mode */
  "",			/* se - leave standout mode */
  "",			/* us - turn on underline mode */
  "",			/* ue - turn off underline mode */

  "",			/* md - enter double bright mode */
  "",			/* mr - enter reverse video mode */
  "",			/* mb - enter blinking mode */
  "",			/* me - turn off all appearance modes */

  "",			/* ks - make function keys transmit */
  ""			/* ke - make function keys work locally */
};

char *p4_dumbterm_rawkeys[] =		/* Strings sent by function keys */
{
  "\033OP",			/* k1 - function keys 1 - 4 from vt100 */
  "\033OQ",			/* k2 */
  "\033OR",			/* k3 */
  "\033OS",			/* k4 */
  "\033[15~",			/* k5 - function keys 5 - 10 from xterm */
  "\033[17~",			/* k6 */
  "\033[18~",			/* k7 */
  "\033[19~",			/* k8 */
  "\033[20~",			/* k9 */
  "\033[21~",			/* k0 */

  "\033[23~",			/* F1 - function keys S-F1 thru S-F10 */
  "\033[24~",			/* F2 - stem from Linux console */
  "\033[25~",			/* F3 - or whoknowswhereelse */
  "\033[26~",			/* F4 */
  "\033[28~",			/* F5 */
  "\033[29~",			/* F6 */
  "\033[31~",			/* F7 */
  "\033[32~",			/* F8 */
  "\033[33~",			/* F9 */
  "\033[34~",			/* FA */

  "\033OD",			/* kl - arrow left */
  "\033OC",			/* kr - arrow right */
  "\033OA",			/* ku - arrow up */
  "\033OB",			/* kd - arrow down */

  "\033[1~",			/* kh - home key */
  "\033[4~",			/* kH - home down key (end key) */
  "\033[6~",			/* kN - next page */
  "\033[5~",			/* kP - previous page */

  "\b",				/* kb - backspace key */
  "\033[3~",			/* kD - delete character key */
  NULL,				/* kM - exit insert mode key */
  "\033[2~",			/* kI - insert character key */

  NULL,				/* kA - insert line key */
  NULL,				/* kE - clear end of line key */
  NULL,				/* kL - delete line key */
  NULL,				/* kC - clear screen key */
};

/*
 * 2.1 simply assume the terminal understands vt100 codes
 */

char *p4_vt100_controls[] =	/* Some hardcoded vt100 sequences. */
{
  "\033[%i%d;%dH",		/* cm - cursor move */
  "\033[H",			/* ho - home position */

  "\b",				/* le - cursor left */
  "\033[C",			/* nd - right one column */
  "\033[A",			/* up - up one column */
  "\n",				/* do - down one column */

  "\033[H\033[2J",		/* cl - clear screen and home */
  "\033[J",			/* cd - clear down */
  "\033[K",			/* ce - clear to end of line */
  "\a",				/* bl - bell */

  "\033[P",			/* dc - delete character in line */
  "\033[M",			/* dl - delete line from screen */

  "\033D",			/* sf - scroll screen up */
  "\033M",			/* sr - scroll screen down */

  "\033[7m",			/* so - enter standout mode */
  "\033[m",			/* se - leave standout mode */
  "\033[4m",			/* us - turn on underline mode */
  "\033[m",			/* ue - turn off underline mode */

  "\033[1m",			/* md - enter double bright mode */
  "\033[7m",			/* mr - enter reverse video mode */
  "\033[5m",			/* mb - enter blinking mode */
  "\033[m",			/* me - turn off all appearance modes */

  "\033[?1h\033=",		/* ks - make function keys transmit */
  "\033[?1l\033>"		/* ke - make function keys work locally */
};

char *p4_vt100_rawkeys[] =		/* Strings sent by function keys */
{
  "\033OP",			/* k1 - function keys 1 - 4 from vt100 */
  "\033OQ",			/* k2 */
  "\033OR",			/* k3 */
  "\033OS",			/* k4 */
  "\033[15~",			/* k5 - function keys 5 - 10 from xterm */
  "\033[17~",			/* k6 */
  "\033[18~",			/* k7 */
  "\033[19~",			/* k8 */
  "\033[20~",			/* k9 */
  "\033[21~",			/* k0 */

  "\033[23~",			/* F1 - function keys S-F1 thru S-F10 */
  "\033[24~",			/* F2 - stem from Linux console */
  "\033[25~",			/* F3 - or whoknowswhereelse */
  "\033[26~",			/* F4 */
  "\033[28~",			/* F5 */
  "\033[29~",			/* F6 */
  "\033[31~",			/* F7 */
  "\033[32~",			/* F8 */
  "\033[33~",			/* F9 */
  "\033[34~",			/* FA */

  "\033OD",			/* kl - arrow left */
  "\033OC",			/* kr - arrow right */
  "\033OA",			/* ku - arrow up */
  "\033OB",			/* kd - arrow down */

  "\033[1~",			/* kh - home key */
  "\033[4~",			/* kH - home down key (end key) */
  "\033[6~",			/* kN - next page */
  "\033[5~",			/* kP - previous page */

  "\b",				/* kb - backspace key */
  "\033[3~",			/* kD - delete character key */
  NULL,				/* kM - exit insert mode key */
  "\033[2~",			/* kI - insert character key */

  NULL,				/* kA - insert line key */
  NULL,				/* kE - clear end of line key */
  NULL,				/* kL - delete line key */
  NULL,				/* kC - clear screen key */
};

/** ASSUME_VT100 ( -- )
 * load hardwired VT100-termcap into the terminal-driver
 */
FCode (p4_assume_vt100)
{
  PFE.control_string = p4_vt100_controls;
  PFE.rawkey_string = p4_vt100_rawkeys;
}

/** ASSUME_DUMBTERM ( -- )
 * load hardwired DUMBTERM-termcap into the terminal-driver
 */
FCode (p4_assume_dumbterm)
{
  PFE.control_string = p4_dumbterm_controls;
  PFE.rawkey_string = p4_dumbterm_rawkeys;
}

/* this one is called *before* loading other things in p4_main */
int p4_prepare_terminal ()
{
  if (!PFE.term) PFE.term = &p4_term_ios;

  PFE.on_stop =		p4_system_terminal;
  PFE.on_continue =	p4_interactive_terminal;
  PFE.on_winchg =	p4_query_winsize;
  
  if (PFE.term->rawkey_string) PFE.rawkey_string = PFE.term->rawkey_string;
  else PFE.rawkey_string = p4_dumbterm_rawkeys;

  if (PFE.term->control_string) PFE.control_string = PFE.term->control_string;
  
  if (PFE.term->init) return (*PFE.term->init)(); else return 1;
}

void p4_cleanup_terminal ()
{
  if (PFE.term->fini) (*PFE.term->fini)();
  
  PFE.term = 0;
}

/* ********************************************************************** */
/* more advanced screen control                                           */
/* ********************************************************************** */

/** GOTOXY ( x y -- )
 * move the cursor to the specified position on the screen -
 * this is usually done by sending a corresponding esc-sequence
 * to the terminal. 
 */
FCode (p4_gotoxy)			
{
  p4_gotoxy (SP[1], SP[0]);
  SP += 2;
}

/** ?XY ( -- x y )
 * returns the cursor position on screen, on a real unix system
 * this includes a special call to the screen driver, in remote
 * systems this can be the expected position as seen on the
 * client side's terminal driver.
 */
FCode (p4_question_xy)		
{				
  int x, y;

  p4_wherexy (&x, &y);
  SP -= 2;
  SP[1] = x;
  SP[0] = y;
}

/* these are defined above */
/* They are renamed here to satisfy the macro CO(char *, p4code) */

#define	p4_dot_home_		p4_dot_home
#define	p4_dot_clrscr_		p4_dot_clrscr
#define	p4_dot_clreol_		p4_dot_clreol
#define p4_dot_bell_		p4_dot_bell
#define p4_dot_highlight_	p4_dot_standout_on
#define p4_dot_highlight_off_	p4_dot_standout_off
#define p4_dot_underline_	p4_dot_underline_on
#define p4_dot_underline_off_	p4_dot_underline_off
#define p4_dot_intensity_	p4_dot_bright
#define p4_dot_intensity_off_	p4_dot_normal
#define p4_dot_blink_		p4_dot_blinking
#define p4_dot_blink_off_	p4_dot_normal
#define p4_dot_reverse_		p4_dot_reverse
#define p4_dot_reverse_off_	p4_dot_normal
#define p4_dot_normal_		p4_dot_normal


LISTWORDS (term) =
{
  CO ("SHOW-TERMCAP",		p4_show_termcap),
  CO ("SHOW-TERM-CONTROLS",	p4_show_control_strings),
  CO ("SHOW-TERM-ESC-KEYS",	p4_show_rawkey_strings),
  CO ("ASSUME_VT100", 		p4_assume_vt100),
  CO ("ASSUME_DUMBTERM",	p4_assume_dumbterm),
  
  /** words for more advanced screen control */
  DV ("ROWS",		rows),
  DV ("COLS",		cols),
  DV ("XMAX",		xmax),
  DV ("YMAX",		ymax),
  CO ("GOTOXY",		p4_gotoxy),
  CO ("?XY",		p4_question_xy),
  CO ("CLS",		p4_dot_clrscr),
  CO (".CLRSCR",	p4_dot_clrscr),
  CO (".CLREOL",	p4_dot_clreol),
  CO (".HOME",		p4_dot_home),
  CO (".HIGHLIGHT",	p4_dot_highlight),
  CO (".HIGHLIGHT.OFF",	p4_dot_highlight_off),
  CO (".UNDERLINE",	p4_dot_underline),
  CO (".UNDERLINE.OFF",	p4_dot_underline_off),
  CO (".INTENSITY",	p4_dot_intensity),
  CO (".INTENSITY.OFF",	p4_dot_intensity_off),
  CO (".BLINKING",	p4_dot_blink),
  CO (".BLINKING.OFF",	p4_dot_blink_off),
  CO (".REVERSE",	p4_dot_reverse),
  CO (".REVERSE.OFF",	p4_dot_reverse_off),
  CO (".NORMAL",	p4_dot_normal),
  
  /** ring the bell on the terminal (output the corresponding control) */
  CO (".BELL",		p4_dot_bell),

  /** EKEY code constants for function keys */
  OC ("K-LEFT",		EKEY_kl),
  OC ("K-RIGHT",	EKEY_kr),
  OC ("K-UP",		EKEY_ku),
  OC ("K-DOWN",		EKEY_kd),
  OC ("K-HOME",		EKEY_kh),
  OC ("K-END",		EKEY_kH),
  OC ("K-PRIOR",	EKEY_kP),
  OC ("K-NEXT",		EKEY_kN),
  OC ("K1",		EKEY_k1),
  OC ("K2",		EKEY_k2),
  OC ("K3",		EKEY_k3),
  OC ("K4",		EKEY_k4),
  OC ("K5",		EKEY_k5),
  OC ("K6",		EKEY_k6),
  OC ("K7",		EKEY_k7),
  OC ("K8",		EKEY_k8),
  OC ("K9",		EKEY_k9),
  OC ("K10",		EKEY_k0),
  OC ("S-K1",		EKEY_F1),
  OC ("S-K2",		EKEY_F2),
  OC ("S-K3",		EKEY_F3),
  OC ("S-K4",		EKEY_F4),
  OC ("S-K5",		EKEY_F5),
  OC ("S-K6",		EKEY_F6),
  OC ("S-K7",		EKEY_F7),
  OC ("S-K8",		EKEY_F8),
  OC ("S-K9",		EKEY_F9),
  OC ("S-K10",		EKEY_FA),
};
COUNTWORDS (term, "PFE-TERM terminal interface(s)");

/*@}*/

