/** @name term.h ---	Declarations for the terminal drivers of pfe
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE            @(#) %derived_by: guidod %
 *  @version %version: 4.30 %
 *    (%date_modified: Thu Jan 27 21:21:38 2000 %)
 *
 *  @description
 *		It doesn't matter how you supply these functions,
 *		  --- just do it!
 *  
 *		For an example driver refer to termunix.c.
 */
/*@{*/

#ifndef __P4_TERM_H
#define __P4_TERM_H

enum ekey_codes			/* The function keys known to p4th */
{
  EKEY_k1 = 0x100,		/* function keys F1 through F10 */
  EKEY_k2,
  EKEY_k3,
  EKEY_k4,
  EKEY_k5,
  EKEY_k6,
  EKEY_k7,
  EKEY_k8,
  EKEY_k9,
  EKEY_k0,

  EKEY_F1,			/* function keys S-F1 through S-F10 */
  EKEY_F2,
  EKEY_F3,
  EKEY_F4,
  EKEY_F5,
  EKEY_F6,
  EKEY_F7,
  EKEY_F8,
  EKEY_F9,
  EKEY_FA,

  /* "kl", "kr", "ku", "kd" */
  EKEY_kl,			/* arrow key left */
  EKEY_kr,			/* arrow key right */
  EKEY_ku,			/* arrow key up */
  EKEY_kd,			/* arrow key down */
  /* "kh", "kH", "kN", "kP" */
  EKEY_kh,			/* home key */
  EKEY_kH,			/* home down key */
  EKEY_kN,			/* next page key */
  EKEY_kP,			/* previous page key */
  /* "kb", "kD", "kM", "kI" */
  EKEY_kb,			/* backspace key */
  EKEY_kD,			/* delete character key */
  EKEY_kM,			/* exit insert mode key */
  EKEY_kI,			/* insert character key */
  /* "kA", "kE", "kL", "kC" */
  EKEY_kA,			/* insert line key */
  EKEY_kE,			/* clear to end of line key */
  EKEY_kL,			/* delete line key */
  EKEY_kC,			/* clear screen key */
  /* count the keys: */
  EKEY_LAST,
  NO_OF_KEYS = EKEY_LAST - EKEY_k1
};

struct _p4_term_struct
{
  char* name;
  char** control_string;
  char** rawkey_string;
  int  (*init) (void);
  void (*fini) (void);

  void (*tput)(int);

  int (*tty_interrupt_key) (char ch);
  void (*interactive_terminal) (void);
  void (*system_terminal) (void);
  void (*query_winsize) (void);

   int (*c_keypressed) (void);	
   int (*c_getkey) (void);

   void (*c_putc_noflush) (char c);
   void (*c_put_flush) (void);
   void (*c_putc) (char c);
   void (*c_puts) (const char *s);
   void (*c_gotoxy) (int x, int y);
   void (*c_wherexy) (int *x, int *y);
};

extern struct _p4_term_struct p4_term_ios;

/* DELETE
extern char** rawkey_string;	/ * what all those keys really send * /
*/

/*
 * These variables are defined in the term.c and initialized by
 * term-xxx:prepare_terminal().  If window size can change, it would
 * be nice if they were kept up to date.
 */
/* DELETE
extern int rows, cols;		/ * size of text screen * /
extern int xmax, ymax;		/ * size of graphics window in pixels * /
*/

/* Refer to termcap.c for what these functions are expected to do: */

int p4_tty_interrupt_key (char ch);
int p4_prepare_terminal (void);
void p4_interactive_terminal (void);
void p4_system_terminal (void);
void p4_query_winsize (void);

int p4_keypressed (void);	
int p4_getkey (void);

void p4_putc_noflush (char c);
void p4_put_flush (void);
void p4_putc (char c);
void p4_puts (const char *s);
void p4_gotoxy (int x, int y);
void p4_wherexy (int *x, int *y);

void p4_goleft (void);
void p4_goright (void);
void p4_goup (void);
void p4_godown (void);

void p4_dot_home (void);
void p4_dot_clrscr (void);
void p4_dot_clreol (void);
void p4_dot_clrdown (void);
void p4_dot_bell (void);

void p4_dot_standout_on (void);
void p4_dot_standout_off (void);
void p4_dot_bright (void);
void p4_dot_reverse (void);
void p4_dot_blinking (void);
void p4_dot_normal (void);
void p4_dot_underline_on (void);
void p4_dot_underline_off (void);

/* These are not part of the driver, but system independent, in term.c */

int  p4_isprint (int c);	/* like isprint() for ISO-characters */
void p4_putc_printable (int c);	/* like p4_putc() but certainly visible */
int  p4_change_case (int key);	/* exchange lower case with upper case char */
int  p4_getekey (void);		/* get a character like EKEY */
int  p4_ekeypressed (void);	/* check for extended key available */
int  p4_getwskey (void);	/* get a character, for block editor */

extern void (*on_stop) (void);
extern void (*on_continue) (void);
extern void (*on_winchg) (void);

extern void (*p4_ptr_wait_for_stdin)();

/* these are modelled after the tc-codes in termunix ... */
enum p4tcap
{
  P4_TERM_XY = 0, /* not used */
  P4_TERM_HOME,
  
  P4_TERM_GOLEFT,
  P4_TERM_GORIGHT,
  P4_TERM_GOUP,
  P4_TERM_GODOWN,
  
  P4_TERM_CLRSCR,
  P4_TERM_CLRDOWN,
  P4_TERM_CLREOL,

  P4_TERM_BELL,
  P4_TERM_DELCH, /* not used */
  P4_TERM_DELLN, /* not used */
  
  P4_TERM_SCROLLDOWN, /* not used */
  P4_TERM_SCROLLUP,   /* not used */
  
  P4_TERM_BOLD_ON,
  P4_TERM_BOLD_OFF,
  P4_TERM_UNDERLINE_ON,
  P4_TERM_UNDERLINE_OFF,
  P4_TERM_BRIGHT,
  P4_TERM_REVERSE,
  P4_TERM_BLINKING,
  P4_TERM_NORMAL,
  
  P4_TERM_KEYPAD_ON,  /* not used */
  P4_TERM_KEYPAD_OFF, /* not used */

  P4_TERM_MAXENUM
};

#endif
  /* ndef __P4_TERM_H */
/*@}*/
