/** @name EXCEPTION ---    The Optional Exception Word Set
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.26 %
 *    (%date_modified: Thu Jan 27 20:46:36 2000 %)
 *
 *  @description
 *      These words implement an exception system in the
 *      widely known => THROW &amp; => CATCH concept.
 *
 *      see the PFE-SIG wordset for catching OS traps.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: xception.c~4.26:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "support.h"

#include <string.h>
#include <ctype.h>

#include "missing.h"

/** CATCH ( xt -- 0|n )
 * execute the given execution-token and catch
 * any exception that can be caught therein.
 * software can arbitrarily raise an exception
 * using => THROW - the value 0 means there
 * was no exception, other denote implementation
 * dependent exception-codes.
 */
FCode (p4_catch)
{
  p4cell catch_code = p4_catch ((p4xt) *SP++);
  FX_PUSH (catch_code);
}

/** THROW ( n -- )
 * raise an exception - it will adjust the depth
 * of all stacks and start interpreting at the point 
 * of the latest => CATCH <br>
 * if n is null nothing happens, the -1 (ie. => FALSE )
 * is the raise-code of => ABORT - the other codes
 * are implementation dependent and will result in
 * something quite like => ABORT
 */
FCode (p4_throw)
{
  p4cell n = FX_POP;

  switch (n)
    {
    case 0:
      return;
    case 2:
      p4_throw (2, (char *) SP[1], SP[0]);
    default:
      p4_throw (n);
    }
}

LISTWORDS (exception) =
{
  CO ("CATCH", p4_catch),
  CO ("THROW", p4_throw),
};
COUNTWORDS (exception, "EXCEPTION-EXT Exception + extensions");

/*@}*/

